# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import pickle

import pytest
from numpy.testing import assert_equal

from MDAnalysis.lib.util import anyopen
from MDAnalysis.lib.picklable_file_io import (
    BufferIOPicklable,
    FileIOPicklable,
    TextIOPicklable,
    BZ2Picklable,
    GzipPicklable,
    pickle_open,
    bz2_pickle_open,
    gzip_pickle_open,
)
from MDAnalysis.coordinates.GSD import GSDPicklable, gsd_pickle_open, HAS_GSD
from MDAnalysis.coordinates.TRJ import (
    NCDFPicklable,
)
from MDAnalysis.coordinates.chemfiles import check_chemfiles_version

if check_chemfiles_version():
    from MDAnalysis.coordinates.chemfiles import ChemfilesPicklable
from MDAnalysis.coordinates.H5MD import HAS_H5PY
from MDAnalysis.coordinates.H5MD import H5PYPicklable

from MDAnalysis.tests.datafiles import (
    PDB,
    XYZ,
    XYZ_bz2,
    MMTF_gz,
    GMS_ASYMOPT,
    GSD,
    NCDF,
    TPR_xvf,
    H5MD_xvf,
)


@pytest.fixture(
    params=[
        # filename mode
        (PDB, "r"),
        (PDB, "rt"),
        (XYZ_bz2, "rt"),
        (GMS_ASYMOPT, "rt"),
    ]
)
def f_text(request):
    filename, mode = request.param
    return anyopen(filename, mode)


def test_get_right_open_handler_text(f_text):
    assert_equal(f_text.__class__, TextIOPicklable)


def test_iopickle_text(f_text):
    f_text_pickled = pickle.loads(pickle.dumps(f_text))
    assert_equal(f_text.readlines(), f_text_pickled.readlines())


def test_offset_text_same(f_text):
    f_text.readline()
    f_text_pickled = pickle.loads(pickle.dumps(f_text))
    assert_equal(f_text_pickled.tell(), f_text.tell())


@pytest.fixture(
    params=[
        # filename mode ref_class
        (PDB, "rb", BufferIOPicklable),
        (XYZ_bz2, "rb", BZ2Picklable),
        (MMTF_gz, "rb", GzipPicklable),
    ]
)
def f_byte(request):
    filename, mode, ref_reader_class = request.param
    return anyopen(filename, mode), ref_reader_class


def test_get_right_open_handler_byte(f_byte):
    assert_equal(f_byte[0].__class__, f_byte[1])


def test_iopickle_byte(f_byte):
    file = f_byte[0]
    f_byte_pickled = pickle.loads(pickle.dumps(file))
    assert_equal(file.readlines(), f_byte_pickled.readlines())


def test_offset_byte_to_tell(f_byte):
    file = f_byte[0]
    file.readline()
    f_byte_pickled = pickle.loads(pickle.dumps(file))
    assert_equal(f_byte_pickled.tell(), file.tell())


def test_context_manager_pickle():
    with pickle_open(PDB) as file:
        file_pickled = pickle.loads(pickle.dumps(file))
        assert_equal(file.readlines(), file_pickled.readlines())


def test_fileio_pickle():
    raw_io = FileIOPicklable(PDB)
    raw_io_pickled = pickle.loads(pickle.dumps(raw_io))
    assert_equal(raw_io.readlines(), raw_io_pickled.readlines())


@pytest.fixture(
    params=[
        # filename mode open_func open_class
        ("test.pdb", "w", pickle_open, FileIOPicklable),
        ("test.pdb", "x", pickle_open, FileIOPicklable),
        ("test.pdb", "a", pickle_open, FileIOPicklable),
        ("test.bz2", "w", bz2_pickle_open, BZ2Picklable),
        ("test.gz", "w", gzip_pickle_open, GzipPicklable),
    ]
)
def unpicklable_f(request):
    filename, mode, open_func, open_class = request.param
    return filename, mode, open_func, open_class


def test_unpicklable_open_mode(unpicklable_f, tmpdir):
    filename, mode, open_func, open_class = unpicklable_f
    with pytest.raises(ValueError, match=r"Only read mode"):
        open_func(tmpdir.mkdir("pickle").join(filename), mode)


def test_pickle_with_write_mode(unpicklable_f, tmpdir):
    filename, mode, open_func, open_class = unpicklable_f
    f_open_by_class = open_class(tmpdir.mkdir("pickle").join(filename), mode)
    with pytest.raises(RuntimeError, match=r"Can only pickle"):
        f_pickled = pickle.loads(pickle.dumps(f_open_by_class))


@pytest.mark.skipif(not HAS_GSD, reason="gsd not installed")
def test_GSD_pickle():
    gsd_io = gsd_pickle_open(GSD, mode="r")
    gsd_io_pickled = pickle.loads(pickle.dumps(gsd_io))
    assert_equal(
        gsd_io[0].particles.position, gsd_io_pickled[0].particles.position
    )


@pytest.mark.skipif(not HAS_GSD, reason="gsd not installed")
def test_GSD_with_write_mode(tmpdir):
    with pytest.raises(ValueError, match=r"Only read mode"):
        gsd_io = gsd_pickle_open(tmpdir.mkdir("gsd").join("t.gsd"), mode="w")


def test_NCDF_pickle():
    ncdf_io = NCDFPicklable(NCDF, mmap=None)
    ncdf_io_pickled = pickle.loads(pickle.dumps(ncdf_io))
    assert_equal(
        ncdf_io.variables["coordinates"][0],
        ncdf_io_pickled.variables["coordinates"][0],
    )


def test_NCDF_mmap_pickle():
    ncdf_io = NCDFPicklable(NCDF, mmap=False)
    ncdf_io_pickled = pickle.loads(pickle.dumps(ncdf_io))
    assert_equal(ncdf_io_pickled.use_mmap, False)


@pytest.mark.skipif(
    not check_chemfiles_version(), reason="Wrong version of chemfiles"
)
def test_Chemfiles_pickle():
    chemfiles_io = ChemfilesPicklable(XYZ)
    chemfiles_io_pickled = pickle.loads(pickle.dumps(chemfiles_io))
    #  frame has to be first saved to get the right position value.
    #  As opposed to `chemfiles_io.read().positions)
    frame = chemfiles_io.read()
    frame_pickled = chemfiles_io_pickled.read()
    assert_equal(frame.positions[:], frame_pickled.positions[:])


@pytest.mark.skipif(
    not check_chemfiles_version(), reason="Wrong version of chemfiles"
)
def test_Chemfiles_with_write_mode(tmpdir):
    with pytest.raises(ValueError, match=r"Only read mode"):
        chemfiles_io = ChemfilesPicklable(
            tmpdir.mkdir("xyz").join("t.xyz"), mode="w"
        )


@pytest.mark.skipif(not HAS_H5PY, reason="h5py not installed")
def test_H5MD_pickle():
    h5md_io = H5PYPicklable(H5MD_xvf, "r")
    h5md_io_pickled = pickle.loads(pickle.dumps(h5md_io))
    assert_equal(
        h5md_io["particles/trajectory/position/value"][0],
        h5md_io_pickled["particles/trajectory/position/value"][0],
    )


@pytest.mark.skipif(not HAS_H5PY, reason="h5py not installed")
def test_H5MD_pickle_with_driver():
    h5md_io = H5PYPicklable(H5MD_xvf, "r", driver="core")
    h5md_io_pickled = pickle.loads(pickle.dumps(h5md_io))
    assert_equal(
        h5md_io["particles/trajectory/position/value"][0],
        h5md_io_pickled["particles/trajectory/position/value"][0],
    )



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#

import MDAnalysis as mda

import pytest

from pytest import approx

from MDAnalysis.analysis.nucleicacids import (
    NucPairDist,
    WatsonCrickDist,
    MajorPairDist,
    MinorPairDist,
)

from MDAnalysisTests.datafiles import RNA_PSF, RNA_PDB

from MDAnalysis.core.groups import ResidueGroup


@pytest.fixture(scope="module")
def u():
    return mda.Universe(RNA_PSF, RNA_PDB)


@pytest.fixture(scope="module")
def strand(unv=u):
    unv = mda.Universe(RNA_PSF, RNA_PDB)
    return unv.select_atoms("segid RNAA")


def test_empty_ag_error(strand):
    strand1 = ResidueGroup([strand.residues[0]])
    strand2 = ResidueGroup([strand.residues[1]])

    with pytest.raises(ValueError, match="returns an empty AtomGroup"):
        NucPairDist.select_strand_atoms(strand1, strand2, "UNK1", "O2")


@pytest.fixture(scope="module")
def wc_rna(strand, client_NucPairDist):
    strand1 = ResidueGroup([strand.residues[0], strand.residues[21]])
    strand2 = ResidueGroup([strand.residues[1], strand.residues[22]])

    WC = WatsonCrickDist(strand1, strand2)
    WC.run(**client_NucPairDist)
    return WC


def test_wc_dist_shape(wc_rna):
    assert wc_rna.results.distances.shape == (1, 2)


def test_wc_dist_results_keys(wc_rna):
    assert "distances" in wc_rna.results


def test_wc_dist(wc_rna):
    assert wc_rna.results.distances[0, 0] == approx(4.3874702, rel=1e-3)
    assert wc_rna.results.distances[0, 1] == approx(4.1716404, rel=1e-3)


def test_wc_dist_invalid_residue_types(u):
    strand = u.select_atoms("resid 1-10")
    strand1 = ResidueGroup([strand.residues[0], strand.residues[21]])
    strand2 = ResidueGroup([strand.residues[2], strand.residues[22]])
    with pytest.raises(ValueError, match="is not a valid nucleic acid"):
        WatsonCrickDist(strand1, strand2)


def test_selection_length_mismatch(strand):
    sel1 = strand.residues[1:10]
    sel2 = strand.residues[1:9]
    with pytest.raises(ValueError, match="Selections must be same length"):
        NucPairDist(sel1, sel2)


def test_wc_dist_deprecation_warning(strand):
    strand1 = [strand.residues[0], strand.residues[21]]
    strand2 = [strand.residues[1], strand.residues[22]]

    with pytest.deprecated_call():
        WatsonCrickDist(strand1, strand2)


def test_wc_dist_strand_verification(strand):
    strand1 = [strand.residues[0], strand[0]]
    strand2 = [strand.residues[1], strand.residues[22]]

    with pytest.raises(TypeError, match="contains non-Residue elements"):
        WatsonCrickDist(strand1, strand2)


@pytest.mark.parametrize("key", [0, 1, 2, "parsnips", "time", -1])
def test_wc_dis_results_keyerrs(wc_rna, key):
    with pytest.raises(KeyError, match=f"{key}"):
        wc_rna.results[key]


def test_minor_dist(strand, client_NucPairDist):
    strand1 = ResidueGroup([strand.residues[2], strand.residues[19]])
    strand2 = ResidueGroup([strand.residues[16], strand.residues[4]])

    MI = MinorPairDist(strand1, strand2)
    MI.run(**client_NucPairDist)

    assert MI.results.distances[0, 0] == approx(15.06506, rel=1e-3)
    assert MI.results.distances[0, 1] == approx(3.219116, rel=1e-3)


def test_major_dist(strand, client_NucPairDist):
    strand1 = ResidueGroup([strand.residues[1], strand.residues[4]])
    strand2 = ResidueGroup([strand.residues[11], strand.residues[8]])

    MA = MajorPairDist(strand1, strand2)
    MA.run(**client_NucPairDist)

    assert MA.results.distances[0, 0] == approx(26.884272, rel=1e-3)
    assert MA.results.distances[0, 1] == approx(13.578535, rel=1e-3)



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
"""
Distance Matrix calculation --- :mod:`MDAnalysis.analysis.ensemble.confdistmatrix`
==================================================================================


The module contains a base class to easily compute, using
parallelization and shared memory, matrices of conformational
distance between the structures stored as frames in a Universe. A
class to compute an RMSD matrix in such a way is also available.

:Author: Matteo Tiberti, Wouter Boomsma, Tone Bengtsen

.. versionadded:: 0.16.0

.. deprecated:: 2.8.0
   This module is deprecated in favour of the 
   MDAKit `mdaencore <https://mdanalysis.org/mdaencore/>`_ and will be removed
   in MDAnalysis 3.0.0.

"""
from joblib import Parallel, delayed
import numpy as np
from getpass import getuser
from socket import gethostname
from datetime import datetime
from time import sleep
import logging
import warnings

from ...core.universe import Universe

from ..align import rotation_matrix

from .cutils import PureRMSD
from .utils import TriangularMatrix, trm_indices


def conformational_distance_matrix(
    ensemble,
    conf_dist_function,
    select="",
    superimposition_select="",
    n_jobs=1,
    pairwise_align=True,
    weights="mass",
    metadata=True,
    verbose=False,
    max_nbytes=None,
):
    """
    Run the conformational distance matrix calculation.
    args and kwargs are passed to conf_dist_function.

    Parameters
    ----------
    ensemble : Universe object
        Universe object for which the conformational distance matrix will
        be computed.
    conf_dist_function : function object
        Function that fills the matrix with conformational distance
        values. See set_rmsd_matrix_elements for an example.
    select : str, optional
        use this selection for the calculation of conformational distance
    superimposition_select : str, optional
        use atoms from this selection for fitting instead of those of
        `select`
    pairwise_align : bool, optional
        Whether to perform pairwise alignment between conformations.
        Default is True (do the superimposition)
    weights : str/array_like, optional
       weights to be used for fit. Can be either 'mass' or an array_like
    metadata : bool, optional
        Whether to build a metadata dataset for the calculated matrix.
        Default is True.
    n_jobs : int, optional
        Number of cores to be used for parallel calculation
        Default is 1. -1 uses all available cores
    max_nbytes : str, optional
        Threshold on the size of arrays passed to the workers that triggers automated memory mapping in temp_folder (default is None).
        See https://joblib.readthedocs.io/en/latest/generated/joblib.Parallel.html for detailed documentation.
    verbose : bool, optional
        enable verbose output

    Returns
    -------
    conf_dist_matrix : encore.utils.TriangularMatrix object
        Conformational distance matrix in triangular representation.

    """

    # framesn: number of frames
    framesn = len(
        ensemble.trajectory.timeseries(
            ensemble.select_atoms(select), order="fac"
        )
    )

    # Prepare metadata recarray
    if metadata:
        metadata = np.array(
            [
                (
                    gethostname(),
                    getuser(),
                    str(datetime.now()),
                    ensemble.filename,
                    framesn,
                    pairwise_align,
                    select,
                    weights == "mass",
                )
            ],
            dtype=[
                ("host", object),
                ("user", object),
                ("date", object),
                ("topology file", object),
                ("number of frames", int),
                ("pairwise superimposition", bool),
                ("superimposition subset", object),
                ("mass-weighted", bool),
            ],
        )

    # Prepare alignment subset coordinates as necessary

    rmsd_coordinates = ensemble.trajectory.timeseries(
        ensemble.select_atoms(select), order="fac"
    )

    if pairwise_align:
        if superimposition_select:
            subset_select = superimposition_select
        else:
            subset_select = select

        fitting_coordinates = ensemble.trajectory.timeseries(
            ensemble.select_atoms(subset_select), order="fac"
        )
    else:
        fitting_coordinates = None

    if (
        not isinstance(weights, (list, tuple, np.ndarray))
        and weights == "mass"
    ):
        weights = ensemble.select_atoms(select).masses.astype(np.float64)
        if pairwise_align:
            subset_weights = ensemble.select_atoms(
                subset_select
            ).masses.astype(np.float64)
        else:
            subset_weights = None
    elif weights is None:
        weights = np.ones(
            (
                ensemble.trajectory.timeseries(ensemble.select_atoms(select))[
                    0
                ].shape[0]
            )
        ).astype(np.float64)
        if pairwise_align:
            subset_weights = np.ones((fit_coords[0].shape[0])).astype(
                np.float64
            )
        else:
            subset_weights = None
    else:
        if pairwise_align:
            if len(weights) != 2:
                raise RuntimeError(
                    "used pairwise alignment with custom "
                    "weights. Please provide 2 tuple with "
                    "weights for 'select' and "
                    "'superimposition_select'"
                )
            subset_weights = weights[1]
            weights = weights[0]
        else:
            subset_weights = None

    # Allocate for output matrix
    matsize = framesn * (framesn + 1) // 2
    distmat = np.empty(matsize, np.float64)

    # Initialize workers. Simple worker doesn't perform fitting,
    # fitter worker does.
    indices = trm_indices((0, 0), (framesn - 1, framesn - 1))
    Parallel(
        n_jobs=n_jobs,
        verbose=verbose,
        require="sharedmem",
        max_nbytes=max_nbytes,
    )(
        delayed(conf_dist_function)(
            np.int64(element),
            rmsd_coordinates,
            distmat,
            weights,
            fitting_coordinates,
            subset_weights,
        )
        for element in indices
    )

    # When the workers have finished, return a TriangularMatrix object
    return TriangularMatrix(distmat, metadata=metadata)


def set_rmsd_matrix_elements(
    tasks,
    coords,
    rmsdmat,
    weights,
    fit_coords=None,
    fit_weights=None,
    *args,
    **kwargs,
):
    """
    RMSD Matrix calculator

    Parameters
    ----------
    tasks : iterator of int of length 2
        Given a triangular matrix, this function will calculate RMSD
        values from element tasks[0] to tasks[1]. Since the matrix
        is triangular, the trm_indices matrix automatically
        calculates the corrisponding i,j matrix indices.
        The matrix is written as an array in a row-major
        order (see the TriangularMatrix class for details).

        If fit_coords and fit_weights are specified, the structures
        will be superimposed before calculating RMSD, and fit_coords and fit_weights
        will be used to place both structures at their center of mass and
        compute the rotation matrix. In this case, both fit_coords and fit_weights
        must be specified.
    coords : numpy.array
        Array of the ensemble coordinates
    weights : numpy.array
        Array of atomic weights, having the same order as the
        coordinates array
    rmsdmat : encore.utils.TriangularMatrix
        Memory-shared triangular matrix object
    fit_coords : numpy.array or None, optional
        Array of the coordinates used for fitting
    fit_weights : numpy.array. optional
        Array of atomic weights, having the same order as the
        fit_coords array
    """
    i, j = tasks

    if fit_coords is None and fit_weights is None:
        sumweights = np.sum(weights)
        rmsdmat[(i + 1) * i // 2 + j] = PureRMSD(
            coords[i].astype(np.float64),
            coords[j].astype(np.float64),
            coords[j].shape[0],
            weights,
            sumweights,
        )

    elif fit_coords is not None and fit_weights is not None:
        sumweights = np.sum(weights)
        subset_weights = np.asarray(fit_weights) / np.mean(fit_weights)
        com_i = np.average(fit_coords[i], axis=0, weights=fit_weights)
        translated_i = coords[i] - com_i
        subset1_coords = fit_coords[i] - com_i
        com_j = np.average(fit_coords[j], axis=0, weights=fit_weights)
        translated_j = coords[j] - com_j
        subset2_coords = fit_coords[j] - com_j
        rotamat = rotation_matrix(
            subset1_coords, subset2_coords, subset_weights
        )[0]
        rotated_i = np.transpose(np.dot(rotamat, np.transpose(translated_i)))
        rmsdmat[(i + 1) * i // 2 + j] = PureRMSD(
            rotated_i.astype(np.float64),
            translated_j.astype(np.float64),
            coords[j].shape[0],
            weights,
            sumweights,
        )
    else:
        raise TypeError(
            "Both fit_coords and fit_weights must be specified "
            "if one of them is given"
        )


def get_distance_matrix(
    ensemble,
    select="name CA",
    load_matrix=None,
    save_matrix=None,
    superimpose=True,
    superimposition_subset="name CA",
    weights="mass",
    n_jobs=1,
    max_nbytes=None,
    verbose=False,
    *conf_dist_args,
    **conf_dist_kwargs,
):
    """
    Retrieves or calculates the conformational distance (RMSD)
    matrix. The distance matrix is calculated between all the frames of all
    the :class:`~MDAnalysis.core.universe.Universe` objects given as input.
    The order of the matrix elements depends on the order of the coordinates
    of the ensembles and on the order of the input ensembles themselves,
    therefore the order of the input list is significant.

    The distance matrix can either be calculated from input ensembles or
    loaded from an input numpy binary file.

    Please notice that the .npz file does not contain a bi-dimensional array,
    but a flattened representation that is meant to represent the elements of
    an encore.utils.TriangularMatrix object.


    Parameters
    ----------
    ensemble : Universe
    select : str
        Atom selection string in the MDAnalysis format. Default is "name CA"
    load_matrix : str, optional
        Load similarity/dissimilarity matrix from numpy binary file instead
        of calculating it (default is None). A filename is required.
    save_matrix : bool, optional
        Save calculated matrix as numpy binary file (default is None). A
        filename is required.
    superimpose : bool, optional
        Whether to superimpose structures before calculating distance
        (default is True).
    superimposition_subset : str, optional
        Group for superimposition using MDAnalysis selection syntax
        (default is CA atoms: "name CA")
    weights : str/array_like, optional
        weights to be used for fit. Can be either 'mass' or an array_like
    n_jobs : int, optional
        Maximum number of cores to be used (default is 1). If -1 use all cores.
    max_nbytes : str, optional
        Threshold on the size of arrays passed to the workers that triggers automated memory mapping in temp_folder (default is None).
        See https://joblib.readthedocs.io/en/latest/generated/joblib.Parallel.html for detailed documentation.
    verbose : bool, optional
        print progress

    Returns
    -------

    confdistmatrix : encore.utils.TriangularMatrix
        Conformational distance matrix. .
    """

    # Load the matrix if required
    if load_matrix:
        logging.info(
            "        Loading similarity matrix from: {0}".format(load_matrix)
        )
        confdistmatrix = TriangularMatrix(
            size=ensemble.trajectory.timeseries(
                ensemble.select_atoms(select), order="fac"
            ).shape[0],
            loadfile=load_matrix,
        )
        logging.info("        Done!")
        for key in confdistmatrix.metadata.dtype.names:
            logging.info(
                "        {0} : {1}".format(
                    key, str(confdistmatrix.metadata[key][0])
                )
            )

        # Check matrix size for consistency
        if (
            not confdistmatrix.size
            == ensemble.trajectory.timeseries(
                ensemble.select_atoms(select), order="fac"
            ).shape[0]
        ):
            logging.error(
                "ERROR: The size of the loaded matrix and of the ensemble"
                " do not match"
            )
            return None

    # Calculate the matrix
    else:

        # Transfer universe to memory to ensure timeseries() support
        ensemble.transfer_to_memory()

        if (
            not isinstance(weights, (list, tuple, np.ndarray))
            and weights == "mass"
        ):
            weight_type = "Mass"
        elif weights is None:
            weight_type = "None"
        else:
            weight_type = "Custom"
        logging.info(
            "        Perform pairwise alignment: {0}".format(str(superimpose))
        )
        logging.info(
            "        weighted alignment and RMSD: {0}".format(weight_type)
        )
        if superimpose:
            logging.info(
                "        Atoms subset for alignment: {0}".format(
                    superimposition_subset
                )
            )
        logging.info("    Calculating similarity matrix . . .")

        # Use superimposition subset, if necessary. If the pairwise alignment
        # is not required, it will not be performed anyway.
        confdistmatrix = conformational_distance_matrix(
            ensemble,
            conf_dist_function=set_rmsd_matrix_elements,
            select=select,
            pairwise_align=superimpose,
            weights=weights,
            n_jobs=n_jobs,
            max_nbytes=max_nbytes,
            verbose=verbose,
        )

        logging.info("    Done!")

        if save_matrix:
            confdistmatrix.savez(save_matrix)

    return confdistmatrix



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#


"""
Selection exporters
===================

Functions to write a :class:`MDAnalysis.core.groups.AtomGroup` selection
to a file so that it can be used in another programme.

:mod:`MDAnalysis.selections.vmd`
    VMD_ selections
:mod:`MDAnalysis.selections.pymol`
    PyMol_ selections
:mod:`MDAnalysis.selections.gromacs`
    Gromacs_ selections
:mod:`MDAnalysis.selections.charmm`
    CHARMM_ selections

The :class:`MDAnalysis.selections.base.SelectionWriterBase` base class and
helper functions are in :mod:`MDAnalysis.selections.base`, with the
exception of `:func:get_writer`:

.. autofunction:: get_writer
"""
import os.path

from .. import _SELECTION_WRITERS

from . import base
from . import vmd
from . import pymol
from . import gromacs
from . import charmm
from . import jmol


def get_writer(filename: str, defaultformat: str) -> base.SelectionWriterBase:
    """Return a SelectionWriter for `filename` or a `defaultformat`.

    Parameters
    ----------
    filename : str
       name of the output file; the extension is used to guess the file format
    defaultformat : str
       if `filename` does not have an extension, use `defaultformat` instead

    Returns
    -------
    SelectionWriterBase : `type`
        the writer *class* for the detected format

    Raises
    ------
    :exc:`NotImplementedError`
        for any format that is not defined
    """
    format = None
    if filename:
        format = os.path.splitext(filename)[1][1:]  # strip initial dot!
    format = format or defaultformat  # use default if no fmt from fn
    format = format.strip().upper()  # canonical for lookup
    try:
        return _SELECTION_WRITERS[format]
    except KeyError:
        errmsg = (
            f"Writing as {format} is not implemented; only "
            f"{ _SELECTION_WRITERS.keys()} will work."
        )
        raise NotImplementedError(errmsg) from None



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#

"""DL_Poly format reader :mod:`MDAnalysis.coordinates.DLPoly`
=============================================================

Read DL Poly_ format coordinate files

.. _Poly: https://www.sc.stfc.ac.uk/software/type/computational-materials-and-molecular-science/?searchquery=dl_poly
"""
import numpy as np

from . import base
from . import core
from ..lib import util
from ..lib.util import cached, store_init_arguments

_DLPOLY_UNITS = {'length': 'Angstrom', 'velocity': 'Angstrom/ps', 'time': 'ps'}


class ConfigReader(base.SingleFrameReaderBase):
    """DLPoly Config file Reader


    .. versionadded:: 0.11.0
    .. versionchanged:: 2.0.0
       coordinates, velocities, and forces are no longer stored in 'F' memory
       layout, instead now using the numpy default of 'C'.
    """
    format = 'CONFIG'
    units = _DLPOLY_UNITS

    def _read_first_frame(self):
        unitcell = np.zeros((3, 3), dtype=np.float32)

        with open(self.filename, 'r') as inf:
            self.title = inf.readline().strip()
            levcfg, imcon, megatm = np.int64(inf.readline().split()[:3])
            if not imcon == 0:
                unitcell[0][:] = inf.readline().split()
                unitcell[1][:] = inf.readline().split()
                unitcell[2][:] = inf.readline().split()

            ids = []
            coords = []
            if levcfg > 0:
                has_vels = True
                velocities = []
            else:
                has_vels = False
            if levcfg == 2:
                has_forces = True
                forces = []
            else:
                has_forces = False

            line = inf.readline().strip()
            # Read records infinitely
            while line:
                try:
                    idx = int(line[8:])
                except ValueError:  # dl_poly classic doesn't have this
                    pass
                else:
                    ids.append(idx)

                xyz = np.float32(inf.readline().split())
                coords.append(xyz)
                if has_vels:
                    vxyz = np.float32(inf.readline().split())
                    velocities.append(vxyz)
                if has_forces:
                    fxyz = np.float32(inf.readline().split())
                    forces.append(fxyz)

                line = inf.readline().strip()

        coords = np.array(coords, dtype=np.float32)
        if has_vels:
            velocities = np.array(velocities, dtype=np.float32)
        if has_forces:
            forces = np.array(forces, dtype=np.float32)
        self.n_atoms = len(coords)

        if ids:
            # If we have indices then sort based on them
            ids = np.array(ids)
            order = np.argsort(ids)

            coords = coords[order]
            if has_vels:
                velocities = velocities[order]
            if has_forces:
                forces = forces[order]

        ts = self.ts = self._Timestep(self.n_atoms,
                                      velocities=has_vels,
                                      forces=has_forces,
                                      **self._ts_kwargs)
        ts._pos = coords
        if has_vels:
            ts._velocities = velocities
        if has_forces:
            ts._forces = forces
        if not imcon == 0:
            ts.dimensions = core.triclinic_box(*unitcell)

        ts.frame = 0


class HistoryReader(base.ReaderBase):
    """Reads DLPoly format HISTORY files

    .. versionadded:: 0.11.0
    """
    format = 'HISTORY'
    units = _DLPOLY_UNITS

    @store_init_arguments
    def __init__(self, filename, **kwargs):
        super(HistoryReader, self).__init__(filename, **kwargs)
        self._cache = {}

        # "private" file handle
        self._file = util.anyopen(self.filename, 'r')
        self.title = self._file.readline().strip()
        header = np.int64(self._file.readline().split()[:3])
        self._levcfg, self._imcon, self.n_atoms = header
        self._has_vels = True if self._levcfg > 0 else False
        self._has_forces = True if self._levcfg == 2 else False

        rwnd = self._file.tell()
        self._file.readline()
        if (len(self._file.readline().split())) == 3:
            self._has_cell = True
        else:
            self._has_cell = False
        self._file.seek(rwnd)

        self.ts = self._Timestep(self.n_atoms,
                                 velocities=self._has_vels,
                                 forces=self._has_forces,
                                 **self._ts_kwargs)
        self._read_next_timestep()

    def _read_next_timestep(self, ts=None):
        if ts is None:
            ts = self.ts

        line = self._file.readline()  # timestep line
        if not line.startswith('timestep'):
            raise IOError

        if self._has_cell:
            unitcell = np.zeros((3, 3))
            unitcell[0] = self._file.readline().split()
            unitcell[1] = self._file.readline().split()
            unitcell[2] = self._file.readline().split()
            ts.dimensions = core.triclinic_box(*unitcell)            

        # If ids are given, put them in here
        # and later sort by them
        ids = []

        for i in range(self.n_atoms):
            line = self._file.readline().strip()  # atom info line
            try:
                idx = int(line.split()[1])
            except IndexError:
                pass
            else:
                ids.append(idx)

            # Read in this order for now, then later reorder in place
            ts._pos[i] = self._file.readline().split()
            if self._has_vels:
                ts._velocities[i] = self._file.readline().split()
            if self._has_forces:
                ts._forces[i] = self._file.readline().split()
            i += 1

        if ids:
            ids = np.array(ids)
            # if ids aren't strictly sequential
            if not np.all(ids == (np.arange(self.n_atoms) + 1)):
                order = np.argsort(ids)
                ts._pos[:] = ts._pos[order]
                if self._has_vels:
                    ts._velocities[:] = ts._velocities[order]
                if self._has_forces:
                    ts._forces[:] = ts._forces[order]

        ts.frame += 1
        return ts

    def _read_frame(self, frame):
        """frame is 0 based, error checking is done in base.getitem"""
        self._file.seek(self._offsets[frame])
        self.ts.frame = frame - 1  # gets +1'd in read_next_frame
        return self._read_next_timestep()

    @property
    @cached('n_frames')
    def n_frames(self):
        # Second line is traj_key, imcom, n_atoms, n_frames, n_records
        offsets = []

        with open(self.filename, 'r') as f:
            f.readline()
            f.readline()
            position = f.tell()
            line = f.readline()
            while line.startswith('timestep'):
                offsets.append(position)
                if self._has_cell:
                    f.readline()
                    f.readline()
                    f.readline()
                for _ in range(self.n_atoms):
                    f.readline()
                    f.readline()
                    if self._has_vels:
                        f.readline()
                    if self._has_forces:
                        f.readline()
                position = f.tell()
                line = f.readline()

        self._offsets = offsets
        return len(self._offsets)

    def _reopen(self):
        self.close()
        self._file = open(self.filename, 'r')
        self._file.readline()  # header is 2 lines
        self._file.readline()
        self.ts.frame = -1

    def close(self):
        self._file.close()



# Adapted from mdtraj/devtools/travis-ci/update_versions_json.py,
# by Robert T McGibbon, under the LGPLv2.1 license
# Similarly by the terms of LGPL this is also in the public domain
# Lily Wang, 2020
#
# This script is called by deploy_docs_via_travis.sh to:
#  1. update versions.json
#  2. Write some redirect stubs
#  3. Write a sitemap.xml file for the root directory
#

import errno
import glob
import json
import os
import shutil
import textwrap
import xml.etree.ElementTree as ET

try:
    from urllib.request import Request, urlopen
except ImportError:
    from urllib2 import Request, urlopen

URL = os.environ["URL"]
VERSION = os.environ["VERSION"]

if "http" not in URL:
    raise ValueError(
        "URL should have the transfer protocol (HTTP/S). " f"Given: $URL={URL}"
    )

try:
    int(VERSION[0])
except ValueError:
    raise ValueError(
        "$VERSION should start with a number. " f"Given: $VERSION={VERSION}"
    ) from None


def get_web_file(filename, callback, default):
    url = os.path.join(URL, filename)
    try:
        page = Request(url, headers={"User-Agent": "Mozilla/5.0"})
        data = urlopen(page).read().decode()
    except Exception as e:
        print(e)
        try:
            with open(filename, "r") as f:
                return callback(f)
        except IOError as e:
            print(e)
            return default
    else:
        return callback(data)


def write_redirect(file, version="", outfile=None):
    if outfile is None:
        outfile = file
    url = os.path.join(URL, version, file)
    REDIRECT = textwrap.dedent(
        f"""
    <!DOCTYPE html>
    <meta charset="utf-8">
    <title>Redirecting to {url}</title>
    <meta http-equiv="refresh" content="0; URL={url}">
    <link rel="canonical" href="{url}">
    """
    )
    with open(outfile, "w") as f:
        f.write(REDIRECT)
    print(f"Wrote redirect from {url} to {outfile}")


# ========= WRITE JSON =========
# Update $root/versions.json with links to the right version
versions = get_web_file("versions.json", json.loads, [])
existing = [item["version"] for item in versions]
already_exists = VERSION in existing
latest = "dev" not in VERSION

if not already_exists:
    if latest:
        for ver in versions:
            ver["latest"] = False

    versions.append(
        {
            "version": VERSION,
            "display": VERSION,
            "url": os.path.join(URL, VERSION),
            "latest": latest,
        }
    )

for ver in versions[::-1]:
    if ver["latest"]:
        latest_version = ver["version"]
        break
else:
    try:
        latest_version = versions[-1]["version"]
    except IndexError:
        latest_version = None

for ver in versions[::-1]:
    if "-dev" in ver["version"]:
        dev_version = ver["version"]
        break
else:
    try:
        dev_version = versions[-1]["version"]
    except IndexError:
        dev_version = None

versions.sort(key=lambda x: x["version"])

# ========= WRITE HTML STUBS AND COPY DOCS =========
# Add HTML files to redirect:
# index.html -> stable/ docs
# latest/index.html -> latest release (not dev docs)
# stable/ : a copy of the release docs with the highest number (2.0.0 instead of 1.0.0)
# dev/ : a copy of the develop docs with the highest number (2.0.0-dev instead of 1.0.1-dev)
# sitemap.xml files are updated by replacing URL strings


def redirect_sitemap(old_version, new_version):
    """Replace paths in copied sitemap.xml with new directory path

    Sitemaps can only contain URLs 'within' that directory structure.
    For more, see https://www.sitemaps.org/faq.html#faq_sitemap_location
    """
    file = f"{new_version}/sitemap.xml"
    old = f"{URL}/{old_version}/"
    new = f"{URL}/{new_version}/"
    try:
        with open(file, "r") as f:
            contents = f.read()
    except OSError:
        raise ValueError(f"{file} not found")
    redirected = contents.replace(old, new)
    with open(file, "w") as f:
        f.write(redirected)
    print(f"Redirected URLs in {file} from {old} to {new}")


def add_or_update_version(version):
    """Add or update the version path to versions.json"""
    for ver in versions:
        if ver["version"] == version:
            ver["url"] = os.path.join(URL, version)
            break
    else:
        versions.append(
            {
                "version": version,
                "display": version,
                "url": os.path.join(URL, version),
                "latest": False,
            }
        )


def copy_version(old_version, new_version):
    """Copy docs from one directory to another with all bells and whistles"""
    shutil.copytree(old_version, new_version)
    print(f"Copied {old_version} to {new_version}")
    redirect_sitemap(old_version, new_version)
    add_or_update_version(new_version)


# Copy stable/ docs and write redirects from root level docs
if latest:
    copy_version(VERSION, "stable")
    html_files = glob.glob(f"stable/**/*.html", recursive=True)
    for file in html_files:
        # below should be true because we only globbed stable/* paths
        assert file.startswith("stable/")
        outfile = file[7:]  # strip "stable/"
        dirname = os.path.dirname(outfile)
        if dirname and not os.path.exists(dirname):
            try:
                os.makedirs(dirname)
            except OSError as exc:
                if exc.errno != errno.EEXIST:
                    raise

        write_redirect(file, "", outfile)

# Separate just in case we update versions.json or muck around manually
# with docs
if latest_version:
    write_redirect("index.html", "stable")
    write_redirect("index.html", latest_version, "latest/index.html")

# Copy dev/ docs
if dev_version and dev_version == VERSION:
    copy_version(VERSION, "dev")

# update versions.json online
with open("versions.json", "w") as f:
    json.dump(versions, f, indent=2)


# ========= WRITE SUPER SITEMAP.XML =========
# make one big sitemap.xml
ET.register_namespace("xhtml", "http://www.w3.org/1999/xhtml")

# so we could make 1 big sitemap as commented
# below, but they must be max 50 MB / 50k URL.
# Yes, this is 100+ releases, but who knows when
# that'll happen and who'll look at this then?
# bigroot = ET.Element("urlset")
# bigroot.set("xmlns", "http://www.sitemaps.org/schemas/sitemap/0.9")
# for ver in versions:
#     tree = get_web_file(ver['version']+'/sitemap.xml', ET.fromstring,
#                         ET.fromstring(''))
#     root = tree.getroot()
#     bigroot.extend(root.getchildren())
# ET.ElementTree(bigroot).write('sitemap.xml',
#                               xml_declaration=True,
#                               encoding='utf-8',
#                               method="xml")

# so instead we make a sitemap of sitemaps.
bigroot = ET.Element("sitemapindex")
bigroot.set("xmlns", "http://www.sitemaps.org/schemas/sitemap/0.9")
for ver in versions:
    path = os.path.join(URL, "{}/sitemap.xml".format(ver["version"]))
    sitemap = ET.SubElement(bigroot, "sitemap")
    ET.SubElement(sitemap, "loc").text = path

ET.ElementTree(bigroot).write(
    "sitemap_index.xml", xml_declaration=True, encoding="utf-8", method="xml"
)



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#

"""OpenMM topology parser :mod:`MDAnalysis.converters.OpenMMParser`
===================================================================

.. versionadded:: 2.0.0
.. versionchanged:: 2.8.0
   Removed type and mass guessing (attributes guessing takes place
   now through universe.guess_TopologyAttrs() API)


Converts an
`OpenMM topology <http://docs.openmm.org/latest/api-python/generated/openmm.app.topology.Topology.html#openmm.app.topology.Topology>`_
:class:`openmm.app.topology.Topology` into a :class:`MDAnalysis.core.Topology`.

Also converts some objects within the
`OpenMM Application layer <http://docs.openmm.org/latest/api-python/app.html>`_

- `openmm.app.pdbfile.PDBFile <http://docs.openmm.org/latest/api-python/generated/openmm.app.pdbfile.PDBFile.html#openmm.app.pdbfile.PDBFile>`_
- `openmm.app.simulation.Simulation <http://docs.openmm.org/latest/api-python/generated/openmm.app.simulation.Simulation.html#openmm.app.simulation.Simulation>`_
- `openmm.app.modeller.Modeller <http://docs.openmm.org/latest/api-python/generated/openmm.app.modeller.Modeller.html#openmm.app.modeller.Modeller>`_
- `openmm.app.pdbxfile.PDBxFile <http://docs.openmm.org/latest/api-python/generated/openmm.app.pdbxfile.PDBxFile.html#openmm.app.pdbxfile.PDBxFile>`_

The :class:`OpenMMTopologyParser` generates a topology from an OpenMM Topology object.


Classes
-------

.. autoclass:: OpenMMTopologyParser
   :members:
   :inherited-members:

.. autoclass:: OpenMMAppTopologyParser
   :members:
   :inherited-members:

"""

import numpy as np
import warnings

from ..topology.base import TopologyReaderBase
from ..guesser.tables import SYMB2Z
from ..core.topology import Topology
from ..core.topologyattrs import (
    Atomids,
    Atomnames,
    Atomtypes,
    Bonds,
    ChainIDs,
    Elements,
    Masses,
    Resids,
    Resnums,
    Resnames,
    Segids,
)


class OpenMMTopologyParser(TopologyReaderBase):
    format = "OPENMMTOPOLOGY"

    @staticmethod
    def _format_hint(thing):
        """Can this Parser read object *thing*?"""
        try:
            from openmm import app
        except ImportError:
            try:  # pragma: no cover
                from simtk.openmm import app
            except ImportError:
                return False
        else:
            return isinstance(thing, app.Topology)

    def _mda_topology_from_omm_topology(self, omm_topology):
        """Construct mda topology from omm topology

        Can be used for any openmm object that contains a topology object

        Parameters
        ----------
        omm_topology: openmm.Topology

        Returns
        -------
        top : MDAnalysis.core.topology.Topology


        When partial elements are present, values from available elements
        are used whereas the absent elements are assigned an empty string
        with their atomtype set to 'X' and mass set to 0.0.

        For elements with invalid and unreal symbols, the symbol is used
        as it is for atomtypes but an empty string is used for elements.

        .. versionchanged:: 2.2.0
           The parser now works when element attribute is missing in some or
           all the atoms.

        """

        try:
            from openmm.unit import daltons
        except ImportError:
            try:
                from simtk.unit import daltons
            except ImportError:
                msg = (
                    "OpenMM is required for the OpenMMParser but "
                    "it's not installed. Try installing it with \n"
                    "conda install -c conda-forge openmm"
                )
                raise ImportError(msg)

        atom_resindex = [a.residue.index for a in omm_topology.atoms()]
        residue_segindex = [r.chain.index for r in omm_topology.residues()]
        atomids = [a.id for a in omm_topology.atoms()]
        atomnames = [a.name for a in omm_topology.atoms()]
        chainids = [a.residue.chain.id for a in omm_topology.atoms()]
        resnames = [r.name for r in omm_topology.residues()]
        resids = [r.index + 1 for r in omm_topology.residues()]
        resnums = resids.copy()
        segids = [str(c.index) for c in omm_topology.chains()]
        bonds = [(b.atom1.index, b.atom2.index) for b in omm_topology.bonds()]
        bond_orders = [b.order for b in omm_topology.bonds()]
        bond_types = [b.type for b in omm_topology.bonds()]

        attrs = [
            Atomids(np.array(atomids, dtype=np.int32)),
            Atomnames(np.array(atomnames, dtype=object)),
            Bonds(bonds, types=bond_types, order=bond_orders, guessed=False),
            ChainIDs(np.array(chainids, dtype=object)),
            Resids(resids),
            Resnums(resnums),
            Resnames(resnames),
            Segids(segids),
        ]

        validated_elements = []
        masses = []
        atomtypes = []
        for a in omm_topology.atoms():
            elem = a.element
            if elem is not None:
                if elem.symbol.capitalize() in SYMB2Z:
                    validated_elements.append(elem.symbol)
                else:
                    validated_elements.append("")
                atomtypes.append(elem.symbol)
                masses.append(elem.mass.value_in_unit(daltons))
            else:
                validated_elements.append("")
                masses.append(0.0)
                atomtypes.append("X")

        if not all(validated_elements):
            if any(validated_elements):
                warnings.warn(
                    "Element information missing for some atoms. "
                    "These have been given an empty element record "
                )
                if any(i == "X" for i in atomtypes):
                    warnings.warn(
                        "For absent elements, atomtype has been  "
                        "set to 'X' and mass has been set to 0.0. "
                        "If needed these can be guessed using "
                        "universe.guess_TopologyAttrs("
                        "to_guess=['masses', 'types']). "
                        "(for MDAnalysis version 2.x "
                        "this is done automatically,"
                        " but it will be removed in 3.0)."
                    )

                attrs.append(
                    Elements(np.array(validated_elements, dtype=object))
                )

            else:
                wmsg = (
                    "Element information is missing for all the atoms. "
                    "Elements attribute will not be populated. "
                    "Atomtype attribute will be guessed using atom "
                    "name and mass will be guessed using atomtype."
                    "For MDAnalysis version 2.x this is done automatically, "
                    "but it will be removed in MDAnalysis v3.0. "
                    "These can be guessed using "
                    "universe.guess_TopologyAttrs("
                    "to_guess=['masses', 'types']) "
                    "See MDAnalysis.guessers."
                )

                warnings.warn(wmsg)
        else:
            attrs.append(Elements(np.array(validated_elements, dtype=object)))
        attrs.append(Atomtypes(np.array(atomtypes, dtype=object)))
        attrs.append(Masses(np.array(masses)))

        n_atoms = len(atomids)
        n_residues = len(resids)
        n_segments = len(segids)
        top = Topology(
            n_atoms,
            n_residues,
            n_segments,
            attrs=attrs,
            atom_resindex=atom_resindex,
            residue_segindex=residue_segindex,
        )

        return top

    def parse(self, **kwargs):
        omm_topology = self.filename
        top = self._mda_topology_from_omm_topology(omm_topology)

        return top


class OpenMMAppTopologyParser(OpenMMTopologyParser):
    format = "OPENMMAPP"

    @staticmethod
    def _format_hint(thing):
        """Can this Parser read object *thing*?"""
        try:
            from openmm import app
        except ImportError:
            try:  # pragma: no cover
                from simtk.openmm import app
            except ImportError:
                return False
        else:
            return isinstance(
                thing,
                (app.PDBFile, app.Modeller, app.Simulation, app.PDBxFile),
            )

    def parse(self, **kwargs):
        try:
            omm_topology = self.filename.getTopology()
        except AttributeError:
            omm_topology = self.filename.topology
        top = self._mda_topology_from_omm_topology(omm_topology)

        return top



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#

import copy
import warnings
from contextlib import suppress
from io import StringIO
from unittest.mock import Mock

import MDAnalysis as mda
import numpy as np
import pytest
from MDAnalysis.converters.RDKit import (
    RDATTRIBUTES,
    _add_mda_attr_to_rdkit,
    _set_atom_property,
    atomgroup_to_mol,
)
from MDAnalysis.converters.RDKitInferring import (
    MDAnalysisInferrer,
    RDKitInferrer,
    TemplateInferrer,
    reorder_atoms,
    sanitize_mol,
)
from MDAnalysis.guesser.default_guesser import DefaultGuesser
from numpy.testing import assert_allclose, assert_array_equal, assert_equal

from MDAnalysisTests.datafiles import (
    GRO,
    TRR,
    PDB_full,
    PDB_helix,
    mol2_molecule,
)
from MDAnalysisTests.util import import_not_available

with suppress(ImportError):
    from rdkit import Chem
    from rdkit.Chem import AllChem


requires_rdkit = pytest.mark.skipif(
    import_not_available("rdkit"), reason="requires RDKit"
)


@pytest.mark.skipif(
    not import_not_available("rdkit"), reason="only for min dependencies build"
)
class TestRequiresRDKit(object):
    def test_converter_requires_rdkit(self):
        u = mda.Universe(PDB_full)
        with pytest.raises(
            ImportError, match="RDKit is required for the RDKitConverter"
        ):
            u.atoms.convert_to("RDKIT")


@requires_rdkit
class MolFactory:
    @staticmethod
    def mol2_mol():
        return Chem.MolFromMol2File(mol2_molecule, removeHs=False)

    @staticmethod
    def smiles_mol():
        mol = Chem.MolFromSmiles("CN1C=NC2=C1C(=O)N(C(=O)N2C)C")
        mol = Chem.AddHs(mol)
        AllChem.EmbedMultipleConfs(mol, numConfs=3)
        return mol

    @staticmethod
    def dummy_product():
        mol = Chem.RWMol()
        atom = Chem.Atom(1)
        atom.SetIntProp("old_mapno", 0)
        atom.SetUnsignedProp("react_atom_idx", 0)
        mol.AddAtom(atom)
        return mol

    @staticmethod
    def dummy_reactant():
        mol = Chem.RWMol()
        atom = Chem.Atom(1)
        atom.SetIntProp("_MDAnalysis_index", 1)
        mol.AddAtom(atom)
        return mol


@pytest.fixture(scope="function")
def rdmol(request):
    return getattr(MolFactory, request.param)()


@pytest.fixture(scope="function")
def product(request):
    return getattr(MolFactory, request.param)()


def is_isomorphic(mol, ref, useChirality=False):
    return mol.HasSubstructMatch(
        ref, useChirality=useChirality
    ) and ref.HasSubstructMatch(mol, useChirality=useChirality)


@requires_rdkit
class TestRDKitReader(object):
    @pytest.mark.parametrize(
        "rdmol, n_frames",
        [
            ("mol2_mol", 1),
            ("smiles_mol", 3),
        ],
        indirect=["rdmol"],
    )
    def test_coordinates(self, rdmol, n_frames):
        universe = mda.Universe(rdmol)
        assert universe.trajectory.n_frames == n_frames
        expected = np.array(
            [conf.GetPositions() for conf in rdmol.GetConformers()],
            dtype=np.float32,
        )
        assert_equal(expected, universe.trajectory.coordinate_array)

    def test_no_coordinates(self):
        with pytest.warns(UserWarning, match="No coordinates found"):
            u = mda.Universe.from_smiles("CCO", generate_coordinates=False)
        expected = np.empty((1, u.atoms.n_atoms, 3), dtype=np.float32)
        expected[:] = np.nan
        assert_equal(u.trajectory.coordinate_array, expected)

    def test_compare_mol2reader(self):
        universe = mda.Universe(MolFactory.mol2_mol())
        mol2 = mda.Universe(mol2_molecule)
        assert universe.trajectory.n_frames == mol2.trajectory.n_frames
        assert_equal(
            universe.trajectory.ts.positions, mol2.trajectory.ts.positions
        )


@requires_rdkit
class TestRDKitConverter(object):
    @pytest.fixture
    def pdb(self):
        return mda.Universe(PDB_full)

    @pytest.fixture
    def mol2(self):
        u = mda.Universe(mol2_molecule)
        # add elements
        guesser = DefaultGuesser(None)
        elements = np.array(
            [guesser.guess_atom_element(x) for x in u.atoms.types],
            dtype=object,
        )
        u.add_TopologyAttr("elements", elements)
        return u

    @pytest.fixture
    def peptide(self):
        u = mda.Universe(GRO)
        elements = mda.guesser.DefaultGuesser(None).guess_types(u.atoms.names)
        u.add_TopologyAttr("elements", elements)
        return u.select_atoms("resid 2-12")

    @pytest.fixture
    def uo2(self):
        return mda.Universe.from_smiles("O=O")

    def test_no_atoms_attr(self):
        rdkit_converter = mda._CONVERTERS["RDKIT"]().convert
        with pytest.raises(TypeError, match=""):
            rdkit_converter("foo")

    @pytest.mark.parametrize("smi", ["[H]", "C", "O", "[He]"])
    def test_single_atom_mol(self, smi):
        u = mda.Universe.from_smiles(
            smi, addHs=False, generate_coordinates=False
        )
        mol = u.atoms.convert_to.rdkit(inferrer=None)
        assert mol.GetNumAtoms() == 1
        assert mol.GetAtomWithIdx(0).GetSymbol() == smi.strip("[]")

    @pytest.mark.parametrize(
        "resname, n_atoms, n_fragments",
        [
            ("PRO", 14, 1),
            ("ILE", 38, 1),
            ("ALA", 20, 2),
            ("GLY", 21, 3),
        ],
    )
    def test_mol_from_selection(self, peptide, resname, n_atoms, n_fragments):
        mol = peptide.select_atoms("resname %s" % resname).convert_to("RDKIT")
        assert n_atoms == mol.GetNumAtoms()
        assert n_fragments == len(Chem.GetMolFrags(mol))

    @pytest.mark.parametrize(
        "sel_str, atom_index",
        [
            ("resid 1", 0),
            ("resname LYS and name NZ", 1),
            ("resid 34 and altloc B", 2),
        ],
    )
    def test_monomer_info(self, pdb, sel_str, atom_index):
        sel = pdb.select_atoms(sel_str)
        mda_atom = sel.atoms[atom_index]
        umol = sel.convert_to.rdkit(inferrer=None)
        atom = umol.GetAtomWithIdx(atom_index)
        mi = atom.GetMonomerInfo()
        assert mda_atom.altLoc == mi.GetAltLoc()
        assert mda_atom.chainID == mi.GetChainId()
        assert mda_atom.icode == mi.GetInsertionCode()
        assert mda_atom.name == mi.GetName().strip()
        assert mda_atom.occupancy == mi.GetOccupancy()
        assert mda_atom.resname == mi.GetResidueName()
        assert mda_atom.resid == mi.GetResidueNumber()
        assert mda_atom.segindex == mi.GetSegmentNumber()
        assert mda_atom.tempfactor == mi.GetTempFactor()

    @pytest.mark.parametrize(
        "rdmol", ["mol2_mol", "smiles_mol"], indirect=True
    )
    def test_identical_topology(self, rdmol):
        u = mda.Universe(rdmol)
        umol = u.atoms.convert_to("RDKIT")
        assert is_isomorphic(rdmol, umol)
        u2 = mda.Universe(umol)
        assert_equal(u.atoms.bonds, u2.atoms.bonds)
        assert_equal(u.atoms.elements, u2.atoms.elements)
        assert_equal(u.atoms.names, u2.atoms.names)
        assert_allclose(
            u.atoms.positions, u2.atoms.positions, rtol=0, atol=1e-7
        )

    def test_raise_requires_elements(self):
        u = mda.Universe(mol2_molecule)

        # Delete topology attribute (PR #3069)
        u.del_TopologyAttr("elements")

        with pytest.raises(
            AttributeError,
            match="`elements` attribute is required for the RDKitConverter",
        ):
            u.atoms.convert_to("RDKIT")

    def test_warn_guess_bonds(self):
        u = mda.Universe(PDB_helix)
        with pytest.warns(
            UserWarning, match="No `bonds` attribute in this AtomGroup"
        ):
            u.atoms.convert_to("RDKIT")

    def test_bonds_outside_sel(self):
        u = mda.Universe(Chem.MolFromSmiles("CC(=O)C"))
        ag = u.select_atoms("index 1")
        ag.convert_to.rdkit(inferrer=None)

    def test_error_no_hydrogen(self, uo2):
        with pytest.raises(
            AttributeError,
            match="the converter requires all hydrogens to be " "explicit",
        ):
            uo2.atoms.convert_to("RDKIT")

    def test_error_no_hydrogen_implicit(self, uo2):
        with warnings.catch_warnings():
            warnings.simplefilter("error")
            uo2.atoms.convert_to.rdkit(inferrer=None)

    def test_warning_no_hydrogen_force(self, uo2):
        with pytest.warns(
            UserWarning, match="Forcing to continue the conversion"
        ):
            uo2.atoms.convert_to.rdkit(force=True)

    @pytest.mark.parametrize(
        "attr, value, expected",
        [
            ("names", "N", " N  "),
            ("names", "CA", " CA "),
            ("names", "CAT", " CAT"),
            ("names", "N1", " N1 "),
            ("names", "CE2", " CE2"),
            ("names", "C12", " C12"),
            ("names", "HD12", "HD12"),
            ("names", "C123", "C123"),
            ("altLocs", "A", "A"),
            ("chainIDs", "B", "B"),
            ("icodes", "C", "C"),
            ("occupancies", 0.5, 0.5),
            ("resnames", "LIG", "LIG"),
            ("resids", 123, 123),
            ("segindices", 1, 1),
            ("tempfactors", 0.8, 0.8),
        ],
    )
    def test_add_mda_attr_to_rdkit(self, attr, value, expected):
        mi = Chem.AtomPDBResidueInfo()
        _add_mda_attr_to_rdkit(attr, value, mi)
        rdvalue = getattr(mi, "Get%s" % RDATTRIBUTES[attr])()
        assert rdvalue == expected

    @pytest.mark.parametrize("idx", [0, 10, 42])
    def test_other_attributes(self, mol2, idx):
        mol = mol2.atoms.convert_to("RDKIT")
        rdatom = mol.GetAtomWithIdx(idx)
        mda_atom = mol2.atoms[idx]
        assert mda_atom.charge == rdatom.GetDoubleProp("_MDAnalysis_charge")
        assert mda_atom.segid == rdatom.GetProp("_MDAnalysis_segid")
        assert mda_atom.type == rdatom.GetProp("_MDAnalysis_type")

    @pytest.mark.parametrize(
        "sel_str",
        [
            "resname ALA",
            "resname PRO and segid A",
        ],
    )
    def test_index_property(self, pdb, sel_str):
        ag = pdb.select_atoms(sel_str)
        mol = ag.convert_to.rdkit(inferrer=None)
        expected = [i for i in range(len(ag))]
        indices = [a.GetIntProp("_MDAnalysis_index") for a in mol.GetAtoms()]
        assert_equal(indices, expected)

    def test_assign_coordinates(self, pdb):
        mol = pdb.atoms.convert_to.rdkit(inferrer=None)
        positions = mol.GetConformer().GetPositions()
        assert_allclose(positions, pdb.atoms.positions)

    def test_assign_stereochemistry(self, mol2):
        umol = mol2.atoms.convert_to("RDKIT")
        rdmol = Chem.MolFromMol2File(mol2_molecule, removeHs=False)
        assert is_isomorphic(rdmol, umol, useChirality=True)

    def test_trajectory_coords(self):
        u = mda.Universe.from_smiles(
            "CCO", numConfs=3, rdkit_kwargs=dict(randomSeed=42)
        )
        for ts in u.trajectory:
            mol = u.atoms.convert_to("RDKIT")
            positions = mol.GetConformer().GetPositions()
            assert_allclose(positions, ts.positions)

    def test_nan_coords(self):
        u = mda.Universe.from_smiles("CCO")
        xyz = u.atoms.positions
        xyz[0][2] = np.nan
        u.atoms.positions = xyz
        with pytest.warns(UserWarning, match="NaN .* detected"):
            mol = u.atoms.convert_to("RDKIT")
        with pytest.raises(ValueError, match="Bad Conformer Id"):
            mol.GetConformer()

    def test_cache(self):
        """5 tests are performed here:

        * while iterating on timesteps in a trajectory, the number of cached
          objects should not change
        * the cache should remember about 2 atomgroups by default
        * the cache size can be increased
        * the cache is sensitive to arguments passed to the converter as they
          might change the output molecule
        * the cache can be ignored
        """
        cached_func = mda.converters.RDKit.atomgroup_to_mol
        # create universes
        utraj = mda.Universe.from_smiles("CCO", numConfs=5)
        uc = mda.Universe.from_smiles("C")
        ucc = mda.Universe.from_smiles("CC")
        uccc = mda.Universe.from_smiles("CCC")
        # test (1): iterating over frames in a trajectory
        previous_cache = None
        for ts in utraj.trajectory:
            utraj.atoms.convert_to("RDKIT")
            cache = cached_func.cache_info()
            if previous_cache:
                # the cache shouldn't change when iterating on timesteps
                assert cache.currsize == previous_cache.currsize
                previous_cache = copy.deepcopy(cache)
        # test (2): only 2 molecules should be cached by default
        cached_func.cache_clear()
        uc.atoms.convert_to("RDKIT")
        ucc.atoms.convert_to("RDKIT")
        uccc.atoms.convert_to("RDKIT")
        cache = cached_func.cache_info()
        assert cache.currsize == 2
        assert cache.misses == 3
        ucc.atoms.convert_to("RDKIT")  # should be inside of the cache
        assert cached_func.cache_info().hits == 1
        uc.atoms.convert_to("RDKIT")  # outside of the cache
        assert cached_func.cache_info().hits == 1
        assert cached_func.cache_info().misses == 4
        # test (3): increase cache size
        mda.converters.RDKit.set_converter_cache_size(3)
        cached_func = mda.converters.RDKit.atomgroup_to_mol
        assert cached_func.cache_info().maxsize == 3
        # test (4): caching is sensitive to converter arguments
        previous_cache = cached_func.cache_info()
        uc.atoms.convert_to.rdkit(implicit_hydrogens=True)
        cache = cached_func.cache_info()
        assert cache.misses == previous_cache.misses + 1
        # test (5): skip cache
        uc.atoms.convert_to.rdkit(cache=False)
        new_cache = cached_func.cache_info()
        assert cache == new_cache

    def test_sanitize_fail_warning(self):
        params = Chem.SmilesParserParams()
        params.removeHs = False
        params.sanitize = False
        mol = Chem.MolFromSmiles("[H]-N(-[H])(-[H])-[H]", params)
        u = mda.Universe(mol)
        with pytest.warns() as record:
            u.atoms.convert_to.rdkit(inferrer=None)
        assert "Could not sanitize molecule" not in "\n".join(
            [str(r.message) for r in record]
        )

    def test_deprecation_max_iter(self, mol2, monkeypatch):
        mock = Mock(wraps=atomgroup_to_mol)
        monkeypatch.setattr(
            "MDAnalysis.converters.RDKit.atomgroup_to_mol", mock
        )
        with pytest.warns(
            DeprecationWarning, match="Using `max_iter` is deprecated"
        ):
            mol2.atoms.convert_to.rdkit(max_iter=2)
        assert mock.call_args.kwargs["inferrer"].max_iter == 2

    def test_deprecation_NoImplicit(self, mol2, monkeypatch):
        mock = Mock(wraps=atomgroup_to_mol)
        monkeypatch.setattr(
            "MDAnalysis.converters.RDKit.atomgroup_to_mol", mock
        )
        with pytest.warns(
            DeprecationWarning, match="Using `NoImplicit` is deprecated"
        ):
            mol2.atoms.convert_to.rdkit(NoImplicit=False)
        assert mock.call_args.kwargs["inferrer"] is None

    def test_deprecation_atomgroup_to_mol_NoImplicit(self, mol2):
        with pytest.warns(
            DeprecationWarning, match="Using `NoImplicit` is deprecated"
        ):
            atomgroup_to_mol(mol2.atoms, NoImplicit=False)

    def test_atomgroup_to_mol_unexpected_kwargs(self, mol2):
        with pytest.raises(
            ValueError, match="Found unexpected arguments: {'foo': 'bar'}"
        ):
            atomgroup_to_mol(mol2.atoms, foo="bar")

    def test_custom_callable_inferrer(self, mol2):
        def potatoe(mol):
            new = Chem.Mol(mol)
            new.SetProp("_Name", "🥔")
            return new

        mol = mol2.atoms.convert_to.rdkit(inferrer=potatoe)
        assert mol.GetProp("_Name") == "🥔"


@requires_rdkit
class TestRDKitInferringFunctions:
    def test_sanitize_mol_error(self):
        mol = Chem.MolFromSmiles("[NH4]", sanitize=False)
        with pytest.raises(
            Chem.AtomValenceException,
            match="Explicit valence for atom # 0 N, 4, is "
            "greater than permitted",
        ):
            sanitize_mol(mol)

    def test_reorder_atoms_key_error(self):
        mol = Chem.MolFromSmiles("CCO")
        with pytest.warns(
            match="'_MDAnalysis_index' not available on the"
            " input mol atoms, skipping reordering of atoms"
        ):
            reorder_atoms(mol)


class BaseInferrer:
    def add_Hs_remove_bo_and_charges(self, mol):
        """Add hydrogens and remove bond orders and charges from a molecule"""
        mH = Chem.AddHs(mol)
        for atom in mH.GetAtoms():
            atom.SetIsAromatic(False)
            atom.SetFormalCharge(0)
            atom.SetNoImplicit(True)
        for bond in mH.GetBonds():
            bond.SetIsAromatic(False)
            bond.SetBondType(Chem.BondType.SINGLE)
        mH.UpdatePropertyCache(strict=False)
        return mH

    def assert_isomorphic_resonance_structure(self, mol, ref):
        """
        Checks if 2 molecules are isomorphic using their resonance structures
        """
        isomorphic = mol.HasSubstructMatch(ref)
        if not isomorphic:
            isomorphic = bool(
                Chem.ResonanceMolSupplier(mol).GetSubstructMatch(ref)
            )
        assert (
            isomorphic
        ), f"{Chem.MolToSmiles(ref)} != {Chem.MolToSmiles(mol)}"


@requires_rdkit
class TestRDKitMDAnalysisInferrer(BaseInferrer):
    def enumerate_reordered_mol(self, mol):
        """Enumerates all possible starting atoms for a given molecule"""
        # go through each possible starting atom
        for root_atom in mol.GetAtoms():
            smi = Chem.MolToSmiles(mol, rootedAtAtom=root_atom.GetIdx())
            reordered_mol = Chem.MolFromSmiles(smi, sanitize=False)
            for atom in reordered_mol.GetAtoms():
                atom.SetNoImplicit(True)
            reordered_mol.UpdatePropertyCache(strict=False)
            yield reordered_mol

    @pytest.fixture(scope="class")
    def inferrer(self):
        return MDAnalysisInferrer()

    @pytest.mark.parametrize(
        "smi, out",
        [
            ("C(-[H])(-[H])(-[H])-[H]", "C"),
            ("[C](-[H])(-[H])-[C](-[H])-[H]", "C=C"),
            (
                "[C]1(-[H])-[C](-[H])-[C](-[H])-[C](-[H])-[C](-[H])-[C]1(-[H])",
                "c1ccccc1",
            ),
            ("C-[C](-[H])-[O]", "C(=O)C"),
            ("[H]-[C](-[O])-[N](-[H])-[H]", "C(=O)N"),
            ("[N]-[C]-[H]", "N#C"),
            ("C-[C](-[O]-[H])-[O]", "CC(=O)O"),
            ("[P](-[O]-[H])(-[O]-[H])(-[O]-[H])-[O]", "P(O)(O)(O)=O"),
            ("[P](-[O]-[H])(-[O]-[H])(-[O])-[O]", "P([O-])(O)(O)=O"),
            ("[P](-[O]-[H])(-[O])(-[O])-[O]", "P([O-])([O-])(O)=O"),
            ("[P](-[O])(-[O])(-[O])-[O]", "P([O-])([O-])([O-])=O"),
            ("[H]-[O]-[N]-[O]", "ON=O"),
            ("[N]-[C]-[O]", "N#C[O-]"),
        ],
    )
    def test_infer_bond_orders(self, smi, out):
        mol = Chem.MolFromSmiles(smi, sanitize=False)
        mol.UpdatePropertyCache(strict=False)
        inferrer = MDAnalysisInferrer()
        inferrer._infer_bo_and_charges(mol)
        Chem.SanitizeMol(mol)
        mol = Chem.RemoveHs(mol)
        molref = Chem.MolFromSmiles(out)
        assert is_isomorphic(mol, molref), f"{Chem.MolToSmiles(mol)} != {out}"

    @pytest.mark.parametrize(
        "smi, atom_idx, charge",
        [
            ("[C](-[H])(-[H])(-[H])-[O]", 4, -1),
            ("[N]-[C]-[O]", 2, -1),
            ("[N](-[H])(-[H])(-[H])-[H]", 0, 1),
            ("C-[C](-[O])-[O]", 3, -1),
        ],
    )
    def test_infer_charges(self, smi, atom_idx, charge):
        mol = Chem.MolFromSmiles(smi, sanitize=False)
        mol.UpdatePropertyCache(strict=False)
        inferrer = MDAnalysisInferrer()
        inferrer._infer_bo_and_charges(mol)
        Chem.SanitizeMol(mol)
        assert mol.GetAtomWithIdx(atom_idx).GetFormalCharge() == charge

    @pytest.mark.parametrize(
        "smi, out",
        [
            ("[S](-[O]-[H])(-[O]-[H])(-[O])-[O]", "S(=O)(=O)(O)O"),
            ("[S](-[O]-[H])(-[O])(-[O])-[O]", "S(=O)(=O)([O-])O"),
            ("[S](-[O])(-[O])(-[O])-[O]", "S(=O)(=O)([O-])[O-]"),
            (
                "C-[N](-[H])-[C](-[N](-[H])-[H])-[N](-[H])-[H]",
                "CNC(N)=[N+](-[H])-[H]",
            ),
            ("[O]-[C](-[H])-[C](-[H])-[H]", "C([O-])=C"),
            ("C-[N](-[O])-[O]", "C[N+](=O)[O-]"),
            ("C(-[N](-[O])-[O])-[N](-[O])-[O]", "C([N+](=O)[O-])[N+](=O)[O-]"),
            ("C-[N](-[O])-[O].C-[N](-[O])-[O]", "C[N+](=O)[O-].C[N+](=O)[O-]"),
            ("[C-](=O)-C", "[C](=O)-C"),
            ("[H]-[N-]-C", "[H]-[N]-C"),
            (
                "[O]-[C]1-[C](-[H])-[C](-[H])-[C](-[H])-[C](-[H])-[C](-[H])1",
                "[O-]c1ccccc1",
            ),
            (
                "[O]-[C]1-[C](-[H])-[C](-[H])-[C](-[H])-[C]1-[O]",
                "[O-]C1=CC=CC1=O",
            ),
            ("[H]-[C]-[C]-[C](-[H])-[C](-[H])-[H]", "C#CC=C"),
            ("[H]-[C]-[C]-[C]-[C]-[H]", "C#CC#C"),
        ],
    )
    def test_standardize_patterns(self, inferrer, smi, out):
        mol = Chem.MolFromSmiles(smi, sanitize=False)
        mol.UpdatePropertyCache(strict=False)
        mol = inferrer(mol)
        mol = Chem.RemoveHs(mol)
        molref = Chem.MolFromSmiles(out)
        assert is_isomorphic(mol, molref), f"{Chem.MolToSmiles(mol)} != {out}"

    @pytest.mark.parametrize(
        "attr, value, getter",
        [
            ("index", 42, "GetIntProp"),
            ("index", np.int8(42), "GetIntProp"),
            ("index", np.int16(42), "GetIntProp"),
            ("index", np.int32(42), "GetIntProp"),
            ("index", np.int64(42), "GetIntProp"),
            ("index", np.uint8(42), "GetIntProp"),
            ("index", np.uint16(42), "GetIntProp"),
            ("index", np.uint32(42), "GetIntProp"),
            ("index", np.uint64(42), "GetIntProp"),
            ("charge", 4.2, "GetDoubleProp"),
            ("charge", np.float32(4.2), "GetDoubleProp"),
            ("charge", np.float64(4.2), "GetDoubleProp"),
            ("type", "C.3", "GetProp"),
        ],
    )
    def test_set_atom_property(self, attr, value, getter):
        atom = Chem.Atom(1)
        prop = "_MDAnalysis_%s" % attr
        _set_atom_property(atom, prop, value)
        assert getattr(atom, getter)(prop) == value

    def test_ignore_prop(self):
        atom = Chem.Atom(1)
        _set_atom_property(atom, "foo", {"bar": "baz"})
        assert "foo" not in atom.GetPropsAsDict().items()

    @pytest.mark.parametrize(
        "smi",
        [
            "c1ccc(cc1)-c1ccccc1-c1ccccc1",
            "c1cc[nH]c1",
            "O=C([C@H](CC1=C[NH1+]=CN1)[NH3+])[O-]",
        ],
    )
    def test_transfer_properties(self, inferrer, smi):
        mol = Chem.MolFromSmiles(smi)
        mol = self.add_Hs_remove_bo_and_charges(mol)
        old = {}
        for atom in mol.GetAtoms():
            ix = atom.GetIdx()
            atom.SetIntProp("_MDAnalysis_index", ix)
            atom.SetProp("dummy", f"foo_{ix}")
            old[ix] = {"_MDAnalysis_index": ix, "dummy": f"foo_{ix}"}
        newmol = inferrer(mol)
        new = {}
        for a in newmol.GetAtoms():
            ix = a.GetIntProp("_MDAnalysis_index")
            new[ix] = {"_MDAnalysis_index": ix, "dummy": a.GetProp("dummy")}
            props = a.GetPropsAsDict().keys()
            assert "old_mapno" not in props
            assert "react_atom_idx" not in props
        assert new == old

    @pytest.mark.parametrize(
        "smi",
        [
            "c1ccc(cc1)-c1ccccc1-c1ccccc1",
            "c1cc[nH]c1",
            "c1ccc(cc1)-c1ccc(-c2ccccc2)c(-c2ccccc2)c1-c1ccccc1",
            "c1ccc2c(c1)c1ccccc1c1ccccc1c1ccccc1c1ccccc21",
            "c1csc(c1)-c1ccoc1-c1cc[nH]c1",
            "C1=C2C(=NC=N1)N=CN2",
            "CN1C=NC(=C1SC2=NC=NC3=C2NC=N3)[N+](=O)[O-]",
            "c1c[nH]c(c1)-c1ccc(s1)-c1ccoc1-c1c[nH]cc1-c1ccccc1",
            "C=CC=CC=CC=CC=CC=C",
            "NCCCCC([NH3+])C(=O)[O-]",
            "CC(C=CC1=C(C)CCCC1(C)C)=CC=CC(C)=CC=[NH+]C",
            "C#CC=C",
            # HID HIE HIP residues, see PR #2941
            "O=C([C@H](CC1=CNC=N1)N)O",
            "O=C([C@H](CC1=CN=CN1)N)O",
            "O=C([C@H](CC1=C[NH1+]=CN1)[NH3+])[O-]",
            # fixes from PR #3044
            "CCOC(=O)c1cc2cc(C(=O)O)ccc2[nH]1",
            "[O-][n+]1cccnc1",
            "C[n+]1ccccc1",
            "[PH4+]",
            "c1nc[nH]n1",
            "CC(=O)C=C(C)N",
            "CC(C)=CC=C[O-]",
            "O=S(C)(C)=NC",
            "Cc1ccc2c3ncn(Cc4ccco4)c(O)c-3nc2c1",
            "CCCC/C=C/C#CC#CCCCCCCCC(=O)O",
            "c1c2c(=O)n3cccc(C)c3nc2n(C)c(=N)c1C(=O)NCc1cnccc1",
            "N#Cc1c[nH]c(C(=O)NC(=O)c2cc[n+]([O-])cc2)n1",
            "C[C@@H](Oc1cc(F)ccc1Nc1ncnc2cc(N=S3(=O)CCC3)cc(F)c12)C(=O)NCC#N",
            "[O-][n+]1onc2ccccc21",
            "Cc1cc[n+](CC[n+]2ccc(C)cc2)cc1",
            "[O-]c1ccccc1",
            "[O-]C=CC=CCC=CC=[N+](C)C",
            "C=[N+](-[O-])-C",
            "C-[N-]-C(=O)C",
            # amino acids
            "C[C@H](N)C(=O)O",  # A
            "NCC(=O)O",  # G
            "CC[C@H](C)[C@H](N)C(=O)O",  # I
            "CC(C)C[C@H](N)C(=O)O",  # L
            "O=C(O)[C@@H]1CCCN1",  # P
            "CC(C)[C@H](N)C(=O)O",  # V
            "N[C@@H](Cc1ccccc1)C(=O)O",  # F
            "N[C@@H](Cc1c[nH]c2ccccc12)C(=O)O",  # W
            "N[C@@H](Cc1ccc(O)cc1)C(=O)O",  # Y
            "N[C@@H](CC(=O)O)C(=O)O",  # D
            "N[C@@H](CCC(=O)O)C(=O)O",  # E
            "N=C(N)NCCC[C@H](N)C(=O)O",  # R
            "N[C@@H](Cc1c[nH]cn1)C(=O)O",  # H
            "NCCCC[C@H](N)C(=O)O",  # K
            "N[C@@H](CO)C(=O)O",  # S
            "C[C@@H](O)[C@H](N)C(=O)O",  # T
            "N[C@@H](CS)C(=O)O",  # C
            "CSCC[C@H](N)C(=O)O",  # M
            "NC(=O)C[C@H](N)C(=O)O",  # N
            "NC(=O)CC[C@H](N)C(=O)O",  # Q
            # nucleic acids
            "Nc1ncnc2c1ncn2[C@H]1C[C@H](OP(=O)(O)O)[C@@H](COP(=O)(O)O)O1",  # A
            "Cc1cn([C@H]2C[C@H](OP(=O)(O)O)[C@@H](COP(=O)(O)O)O2)c(=O)[nH]c1=O",  # T
            "O=c1ccn([C@H]2C[C@H](OP(=O)(O)O)[C@@H](COP(=O)(O)O)O2)c(=O)[nH]1",  # U
            "Nc1ccn([C@H]2C[C@H](OP(=O)(O)O)[C@@H](COP(=O)(O)O)O2)c(=O)n1",  # C
            "Nc1nc2c(ncn2[C@H]2C[C@H](OP(=O)(O)O)[C@@H](COP(=O)(O)O)O2)c(=O)[nH]1",  # G
            # RDKitConverter benchmark 2022-05-23, fixed by better sorting
            "CC(C)NS(C)(=O)=Nc1ccc(-c2ccc(-c3nc4[nH]c(O[C@@H]5CO[C@@H]6[C@H](O)CO[C@H]56)nc4cc3Cl)cc2)cc1 CHEMBL4111464",
            "C[n+]1c2ccccc2c(C(=O)O)c2[nH]c3ccc(Br)cc3c21 CHEMBL511591",
            "C[n+]1ccccc1-c1ccc(NC(=O)c2ccc(C(=O)Nc3ccc(-c4cccc[n+]4C)cc3)cc2)cc1 CHEMBL3230729",
            "Cc1ccc(/C(O)=C(/C([S-])=NCc2ccccc2)[n+]2cccc(CO)c2)cc1[N+](=O)[O-] CHEMBL1596426",
            "CC(C)(O)c1cc2c(cc1C(C)(C)O)-c1ccccc1-2 CHEMBL1990966",
            "[O-]/N=C1/c2ccccc2-c2nc3ccc(C(F)(F)F)cc3nc21 CHEMBL4557878",
            "N#Cc1c[nH]c(C(=O)Nc2ccc(-c3cccc[n+]3[O-])cc2C2=CCCCC2)n1 CHEMBL1172116",
        ],
    )
    def test_order_independant(self, inferrer, smi):
        # generate mol with hydrogens but without bond orders
        ref = Chem.MolFromSmiles(smi)
        stripped_mol = self.add_Hs_remove_bo_and_charges(ref)
        # enumerate mols with different starting atoms
        for m in self.enumerate_reordered_mol(stripped_mol):
            m = inferrer(m)
            m = Chem.RemoveHs(m)
            self.assert_isomorphic_resonance_structure(m, ref)

    @pytest.mark.xfail(reason="Not currently tackled by the RDKitConverter")
    @pytest.mark.parametrize(
        "smi",
        [
            "C-[N+]#N",
            "C-N=[N+]=[N-]",
            "C-[O+]=C",
            "C-[N+]#[C-]",
        ],
    )
    def test_order_independant_issue_3339(self, smi):
        self.test_order_independant(smi)

    def test_warn_conjugated_max_iter(self):
        smi = "[C-]C=CC=CC=CC=CC=CC=C[C-]"
        mol = Chem.MolFromSmiles(smi)
        inferrer = MDAnalysisInferrer(max_iter=2)
        with pytest.warns(
            UserWarning, match="reasonable number of iterations"
        ):
            inferrer._rebuild_conjugated_bonds(mol)

    def test_deprecation_warning_max_iter(self):
        smi = "c1ccccc1"
        mol = Chem.MolFromSmiles(smi)
        inferrer = MDAnalysisInferrer()
        with pytest.warns(
            DeprecationWarning,
            match="Specifying `max_iter` is deprecated and will" " be removed",
        ):
            inferrer._standardize_patterns(mol, max_iter=1)

    @pytest.mark.parametrize(
        "smi",
        [
            "[Li+]",
            "[Na+]",
            "[K+]",
            "[Rb+]",
            "[Ag+]",
            "[Cs+]",
            "[Mg+2]",
            "[Ca+2]",
            "[Cu+2]",
            "[Zn+2]",
            "[Sr+2]",
            "[Ba+2]",
            "[Al+3]",
            "[Fe+2]",
            "[Cl-]",
            "[O-2]",
            "[Na+].[Cl-]",
        ],
    )
    def test_ions(self, inferrer, smi):
        ref = Chem.MolFromSmiles(smi)
        stripped_mol = self.add_Hs_remove_bo_and_charges(ref)
        mol = inferrer(stripped_mol)
        assert is_isomorphic(mol, ref)

    @pytest.mark.parametrize(
        "smi",
        [
            "O=C([C@H](CC1=C[NH1+]=CN1)[NH3+])[O-]",
            "O=S(C)(C)=NC",
        ],
    )
    def test_reorder_atoms(self, smi):
        mol = Chem.MolFromSmiles(smi)
        mol = Chem.AddHs(mol)
        # remove bond order and charges info
        pdb = Chem.MolToPDBBlock(mol)
        u = mda.Universe(StringIO(pdb), format="PDB")
        # atoms are reordered during infering, and will be reordered back to
        # their original order with Chem.RenumberAtoms
        mu = u.atoms.convert_to.rdkit()
        values = [a.GetSymbol() for a in mu.GetAtoms()]
        expected = [a.GetSymbol() for a in mol.GetAtoms()]
        assert values == expected

    @pytest.mark.parametrize(
        "smi",
        [
            "O=S(C)(C)=NC",
        ],
    )
    def test_warn_empty_coords(self, smi):
        mol = Chem.MolFromSmiles(smi)
        mol = Chem.AddHs(mol)
        # remove bond order and charges info
        pdb = Chem.MolToPDBBlock(mol)
        u = mda.Universe(StringIO(pdb), format="PDB")
        with pytest.warns(match="NaN or empty coordinates detected"):
            u.atoms.convert_to.rdkit()

    def test_pdb_names(self):
        u = mda.Universe(PDB_helix)
        mol = u.atoms.convert_to.rdkit()
        names = u.atoms.names
        rd_names = np.array(
            [a.GetProp("_MDAnalysis_name") for a in mol.GetAtoms()]
        )
        assert (names == rd_names).all()

    @pytest.mark.parametrize(
        "smi",
        [
            r"F/C(Br)=C(Cl)/I",
            r"F\C(Br)=C(Cl)\I",
            "F-C(Br)=C(Cl)-I",
        ],
    )
    def test_bond_stereo_not_stereoany(self, smi):
        u = mda.Universe.from_smiles(smi)
        mol = u.atoms.convert_to.rdkit(force=True)
        for bond in mol.GetBonds():
            if bond.GetBondTypeAsDouble() == 2:
                assert bond.GetStereo() != Chem.BondStereo.STEREOANY

    def test_atom_sorter(self):
        mol = Chem.MolFromSmiles(
            "[H]-[C](-[H])-[C](-[H])-[C]-[C]-[H]", sanitize=False
        )
        # corresponding mol: C=C-C#C
        # atom indices:      1 3 5 6
        mol.UpdatePropertyCache()
        sorted_atoms = sorted(
            [atom for atom in mol.GetAtoms() if atom.GetAtomicNum() > 1],
            key=MDAnalysisInferrer._atom_sorter,
        )
        sorted_indices = [atom.GetIdx() for atom in sorted_atoms]
        assert sorted_indices == [6, 5, 1, 3]

    @pytest.mark.parametrize(
        "sanitize",
        [
            pytest.param(
                True,
                marks=pytest.mark.xfail(
                    reason="Invalid charge/valence", strict=True
                ),
            ),
            False,
        ],
    )
    def test_sanitize(self, sanitize):
        mol = Chem.MolFromSmiles("[H]-[C+](-[H])(-[H])-[H]", sanitize=False)
        mol.UpdatePropertyCache(strict=False)
        inferrer = MDAnalysisInferrer(sanitize=sanitize)
        result = inferrer(mol)
        assert isinstance(result, Chem.Mol)

    @pytest.mark.parametrize(
        ("smi", "expected"),
        [
            ("[NH4+]", [0]),
            ("[NH4]", [-1]),
            ("[NH3]", [0]),
            ("[NH2]", [1]),
            ("[SH2]", [0, 2, 4]),
            ("[SH2](=O)(=O)", [-4, -2, 0]),
        ],
    )
    def test_get_nb_unpaired_electrons(self, smi, expected):
        mol = Chem.MolFromSmiles(smi, sanitize=False)
        mol.UpdatePropertyCache(False)
        atom = mol.GetAtomWithIdx(0)
        assert MDAnalysisInferrer._get_nb_unpaired_electrons(atom) == expected


@requires_rdkit
class TestRDKitTemplateInferrer:
    @pytest.fixture(scope="function")
    def template(self):
        return Chem.MolFromSmiles("[NH3+]-C=O")

    @pytest.fixture(scope="function")
    def template_with_hydrogens(self):
        params = Chem.SmilesParserParams()
        params.removeHs = False
        return Chem.MolFromSmiles("[N+](-[H])(-[H])(-[H])-C(-[H])=O", params)

    @pytest.fixture(scope="function")
    def mol(self, template):
        mol = Chem.AddHs(template)
        AllChem.EmbedMolecule(mol, randomSeed=0xAC1D1C)
        mol.GetAtomWithIdx(0).SetFormalCharge(0)
        mol.UpdatePropertyCache(strict=False)
        for atom in mol.GetAtoms():
            atom.SetIntProp("_MDAnalysis_index", atom.GetIdx())
        return mol

    def test_template_inferring(self, mol, template_with_hydrogens):
        inferrer = TemplateInferrer(
            template=template_with_hydrogens, adjust_hydrogens=False
        )
        new = inferrer(mol)
        assert new.GetAtomWithIdx(0).GetFormalCharge() == 1

    def test_template_inferring_adjust_hydrogens(self, mol, template):
        inferrer = TemplateInferrer(template=template, adjust_hydrogens=True)
        new = inferrer(mol)
        assert new.GetAtomWithIdx(0).GetFormalCharge() == 1

    def test_from_convert_to_method(self, template_with_hydrogens):
        mol = Chem.Mol(template_with_hydrogens)
        AllChem.EmbedMolecule(mol, randomSeed=0xAC1D1C)
        u = mda.Universe(mol)
        mol = u.atoms.convert_to.rdkit(
            inferrer=TemplateInferrer(
                template=template_with_hydrogens, adjust_hydrogens=False
            )
        )
        assert mol.GetAtomWithIdx(0).GetFormalCharge() == 1
        xyz = mol.GetConformer().GetPositions()
        assert_array_equal(xyz, u.atoms.positions)

    def test_adjust_hydrogens_bigger_mol(self):
        u = mda.Universe(GRO, TRR)
        elements = mda.topology.guessers.guess_types(u.atoms.names)
        u.add_TopologyAttr("elements", elements)
        ag = u.select_atoms("resid 1-3")
        inferrer = TemplateInferrer(
            template=Chem.MolFromSmiles(
                "CCC(C)C(C=O)NC(=O)C(CCCNC(N)=[NH2+])NC(=O)C([NH3+])CCSC"
            ),
            adjust_hydrogens=True,
        )
        mol = ag.convert_to.rdkit(inferrer=inferrer)
        xyz = mol.GetConformer().GetPositions()
        assert_array_equal(xyz, ag.positions)


@requires_rdkit
class TestRDKitInferrer(BaseInferrer):
    @pytest.mark.parametrize(
        "smi",
        [
            "c1ccc(cc1)-c1ccccc1-c1ccccc1",
            "c1cc[nH]c1",
            "c1ccc(cc1)-c1ccc(-c2ccccc2)c(-c2ccccc2)c1-c1ccccc1",
            "c1ccc2c(c1)c1ccccc1c1ccccc1c1ccccc1c1ccccc21",
            "c1csc(c1)-c1ccoc1-c1cc[nH]c1",
            "C1=C2C(=NC=N1)N=CN2",
            "c1c[nH]c(c1)-c1ccc(s1)-c1ccoc1-c1c[nH]cc1-c1ccccc1",
            "C=CC=CC=CC=CC=CC=C",
            "NCCCCC([NH3+])C(=O)[O-]",
            "CC(C=CC1=C(C)CCCC1(C)C)=CC=CC(C)=CC=[NH+]C",
            "C#CC=C",
            # amino acids
            "C[C@H](N)C(=O)O",  # A
            "NCC(=O)O",  # G
            "CC[C@H](C)[C@H](N)C(=O)O",  # I
            "CC(C)C[C@H](N)C(=O)O",  # L
            "O=C(O)[C@@H]1CCCN1",  # P
            "CC(C)[C@H](N)C(=O)O",  # V
            "N[C@@H](Cc1ccccc1)C(=O)O",  # F
            "N[C@@H](Cc1c[nH]c2ccccc12)C(=O)O",  # W
            "N[C@@H](Cc1ccc(O)cc1)C(=O)O",  # Y
            "N[C@@H](CC(=O)O)C(=O)O",  # D
            "N[C@@H](CCC(=O)O)C(=O)O",  # E
            "N=C(N)NCCC[C@H](N)C(=O)O",  # R
            "N[C@@H](Cc1c[nH]cn1)C(=O)O",  # H
            "NCCCC[C@H](N)C(=O)O",  # K
            "N[C@@H](CO)C(=O)O",  # S
            "C[C@@H](O)[C@H](N)C(=O)O",  # T
            "N[C@@H](CS)C(=O)O",  # C
            "CSCC[C@H](N)C(=O)O",  # M
            "NC(=O)C[C@H](N)C(=O)O",  # N
            "NC(=O)CC[C@H](N)C(=O)O",  # Q
            # nucleic acids
            "Nc1ncnc2c1ncn2[C@H]1C[C@H](OP(=O)(O)O)[C@@H](COP(=O)(O)O)O1",  # A
            "Cc1cn([C@H]2C[C@H](OP(=O)(O)O)[C@@H](COP(=O)(O)O)O2)c(=O)[nH]c1=O",  # T
            "O=c1ccn([C@H]2C[C@H](OP(=O)(O)O)[C@@H](COP(=O)(O)O)O2)c(=O)[nH]1",  # U
            "Nc1ccn([C@H]2C[C@H](OP(=O)(O)O)[C@@H](COP(=O)(O)O)O2)c(=O)n1",  # C
            "Nc1nc2c(ncn2[C@H]2C[C@H](OP(=O)(O)O)[C@@H](COP(=O)(O)O)O2)c(=O)[nH]1",  # G
        ],
    )
    def test_inferring(self, smi):
        # generate mol with hydrogens but without bond orders
        ref = Chem.MolFromSmiles(smi)
        charge = sum(atom.GetFormalCharge() for atom in ref.GetAtoms())
        inferrer = RDKitInferrer(charge=charge)
        mol = Chem.AddHs(ref)
        AllChem.EmbedMolecule(mol, randomSeed=0xAC1D1C)
        stripped_mol = self.add_Hs_remove_bo_and_charges(mol)
        new = inferrer(stripped_mol)
        new = Chem.RemoveHs(new)
        self.assert_isomorphic_resonance_structure(new, ref)



"""
MDAnalysis topology tables
==========================

.. deprecated:: 2.8.0
   The :mod:`MDAnalysis.topology.tables` module has been moved to
   :mod:`MDAnalysis.guesser.tables`. This import point will
   be removed in release 3.0.0.

The module contains static lookup tables for atom typing etc. The
tables are dictionaries that are indexed by the element.

.. autodata:: atomelements
.. autodata:: masses
.. autodata:: vdwradii

The original raw data are stored as multi-line strings that are
translated into dictionaries with :func:`kv2dict`. In the future,
these tables might be moved into external data files; see
:func:`kv2dict` for explanation of the file format.

.. autofunction:: kv2dict

The raw tables are stored in the strings

.. autodata:: TABLE_ATOMELEMENTS
.. autodata:: TABLE_MASSES
.. autodata:: TABLE_VDWRADII
"""

import warnings
from MDAnalysis.guesser.tables import (
    kv2dict,
    TABLE_ATOMELEMENTS,
    atomelements,
    elements,
    TABLE_MASSES,
    masses,
    TABLE_VDWRADII,
    vdwradii,
    Z2SYMB,
    SYMB2Z,
    SYBYL2SYMB,
)

wmsg = (
    "Deprecated in version 2.8.0\n"
    "MDAnalysis.topology.tables has been moved to "
    "MDAnalysis.guesser.tables. This import point "
    "will be removed in MDAnalysis version 3.0.0"
)
warnings.warn(wmsg, category=DeprecationWarning)



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import bz2
import gzip
import os
import numpy as np
import pytest

import MDAnalysis as mda
from MDAnalysis import NoDataError

from numpy.testing import assert_equal, assert_allclose

from MDAnalysisTests import make_Universe
from MDAnalysisTests.coordinates.reference import (
    RefLAMMPSData,
    RefLAMMPSDataMini,
    RefLAMMPSDataDCD,
    RefLAMMPSDataAdditionalColumns,
)
from MDAnalysisTests.datafiles import (
    LAMMPScnt,
    LAMMPShyd,
    LAMMPSdata,
    LAMMPSdata_mini,
    LAMMPSdata_triclinic,
    LAMMPSDUMP,
    LAMMPSDUMP_allcoords,
    LAMMPSDUMP_nocoords,
    LAMMPSDUMP_triclinic,
    LAMMPSDUMP_image_vf,
    LAMMPS_image_vf,
    LAMMPSdata_additional_columns,
    LAMMPSDUMP_additional_columns,
)


def test_datareader_ValueError():
    from MDAnalysis.coordinates.LAMMPS import DATAReader

    with pytest.raises(ValueError):
        DATAReader("filename")


class _TestLammpsData_Coords(object):
    """Tests using a .data file for loading single frame.

    All topology loading from MDAnalysisTests.data is done in test_topology
    """

    @pytest.fixture(scope="class")
    def u(self):
        return mda.Universe(self.filename)

    def test_n_atoms(self, u):
        assert_equal(u.atoms.n_atoms, self.n_atoms)

    def test_coords(self, u):
        assert_equal(u.atoms[0].position, self.pos_atom1)

    def test_velos(self, u):
        assert_allclose(u.atoms[0].velocity, self.vel_atom1)

    def test_dimensions(self, u):
        assert_equal(u.dimensions, self.dimensions)

    def test_singleframe(self, u):
        with pytest.raises(StopIteration):
            u.trajectory.next()

    def test_seek(self, u):
        with pytest.raises(IndexError):
            u.trajectory[1]

    def test_seek_2(self, u):
        ts = u.trajectory[0]
        assert_equal(type(ts), mda.coordinates.base.Timestep)

    def test_iter(self, u):
        # Check that iterating works, but only gives a single frame
        assert len(list(iter(u.trajectory))) == 1


class TestLammpsData_Coords(_TestLammpsData_Coords, RefLAMMPSData):
    pass


class TestLammpsDataMini_Coords(_TestLammpsData_Coords, RefLAMMPSDataMini):
    pass


@pytest.fixture(
    params=[
        LAMMPSdata,
        LAMMPSdata_mini,
        LAMMPScnt,
        LAMMPShyd,
    ],
    scope="module",
)
def LAMMPSDATAWriter(request, tmpdir_factory):
    filename = request.param
    u = mda.Universe(filename)
    fn = os.path.split(filename)[1]
    outfile = str(tmpdir_factory.mktemp("data").join(fn))

    with mda.Writer(outfile, n_atoms=u.atoms.n_atoms) as w:
        w.write(u.atoms)

    u_new = mda.Universe(outfile)

    return u, u_new


@pytest.fixture(
    params=[
        [LAMMPSdata, True],
        [LAMMPSdata_mini, True],
        [LAMMPScnt, True],
        [LAMMPShyd, True],
        [LAMMPSdata, False],
    ],
    scope="module",
)
def LAMMPSDATAWriter_molecule_tag(request, tmpdir_factory):
    filename, charges = request.param
    u = mda.Universe(filename)
    if not charges:
        u.del_TopologyAttr("charges")

    u.trajectory.ts.data["molecule_tag"] = u.atoms.resids

    fn = os.path.split(filename)[1]
    outfile = str(tmpdir_factory.mktemp("data").join(fn))

    with mda.Writer(outfile, n_atoms=u.atoms.n_atoms) as w:
        w.write(u.atoms)

    u_new = mda.Universe(outfile)

    return u, u_new


def test_unwrap_vel_force():

    u_wrapped = mda.Universe(
        LAMMPS_image_vf, [LAMMPSDUMP_image_vf], format="LAMMPSDUMP"
    )
    u_wrapped.trajectory[-1]

    assert_allclose(
        u_wrapped.atoms.positions[0],
        np.array([2.56616, 6.11565, 7.37956]),
        atol=1e-5,
    )
    assert hasattr(u_wrapped.atoms, "velocities")
    assert hasattr(u_wrapped.atoms, "forces")


def test_unwrap_image_wrap():
    u_unwrapped = mda.Universe(
        LAMMPS_image_vf,
        LAMMPSDUMP_image_vf,
        format="LAMMPSDUMP",
        unwrap_images=True,
    )
    u_unwrapped.trajectory[-1]

    pos = (
        np.array([2.56616, 6.11565, 7.37956])
        + np.array([3, 1, -4]) * u_unwrapped.dimensions[:3]
    )
    assert_allclose(
        u_unwrapped.atoms.positions[0],
        pos,
        atol=1e-5,
    )


def test_unwrap_no_image():
    with pytest.raises(ValueError, match="must have image flag"):
        u_unwrapped = mda.Universe(
            LAMMPSDUMP_allcoords, format="LAMMPSDUMP", unwrap_images=True
        )


class TestLAMMPSDATAWriter(object):
    def test_Writer_dimensions(self, LAMMPSDATAWriter):
        u_ref, u_new = LAMMPSDATAWriter
        assert_allclose(
            u_ref.dimensions,
            u_new.dimensions,
            err_msg="attributes different after writing",
            atol=1e-6,
        )

    def test_Writer_atoms_types(self, LAMMPSDATAWriter):
        u_ref, u_new = LAMMPSDATAWriter
        assert_equal(
            u_ref.atoms.types,
            u_new.atoms.types,
            err_msg="attributes different after writing",
        )

    @pytest.mark.parametrize(
        "attr", ["bonds", "angles", "dihedrals", "impropers"]
    )
    def test_Writer_atoms(self, attr, LAMMPSDATAWriter):
        u_ref, u_new = LAMMPSDATAWriter
        ref = getattr(u_ref.atoms, attr)
        new = getattr(u_new.atoms, attr)
        assert ref == new, "attributes different after writing"

    @pytest.mark.parametrize(
        "attr", ["masses", "charges", "velocities", "positions"]
    )
    def test_Writer_numerical_attrs(self, attr, LAMMPSDATAWriter):
        u_ref, u_new = LAMMPSDATAWriter
        try:
            refvals = getattr(u_ref, attr)
        except AttributeError:
            with pytest.raises(AttributeError):
                getattr(u_new, attr)
        else:
            assert_allclose(
                refvals,
                getattr(u_new.atoms, attr),
                err_msg="attributes different after writing",
                atol=1e-6,
            )


class TestLAMMPSDATAWriter_molecule_tag(object):
    def test_molecule_tag(self, LAMMPSDATAWriter_molecule_tag):
        u_ref, u_new = LAMMPSDATAWriter_molecule_tag
        assert_equal(
            u_ref.atoms.resids,
            u_new.atoms.resids,
            err_msg="resids different after writing",
        )


@pytest.mark.parametrize(
    "filename", ["out.data", "out.data.bz2", "out.data.gz"]
)
def test_datawriter_universe(filename, tmpdir):
    """
    Test roundtrip on datawriter, and also checks compressed files
    can be written (see #4159).
    """
    fn = str(tmpdir.join(filename))

    u = mda.Universe(LAMMPSdata_mini)

    with mda.Writer(fn, n_atoms=len(u.atoms)) as w:
        w.write(u)

    u2 = mda.Universe(fn)

    assert_allclose(u.atoms.positions, u2.atoms.positions)
    assert_allclose(u.dimensions, u2.dimensions)


class TestLAMMPSDATAWriter_data_partial(TestLAMMPSDATAWriter):
    N_kept = 5

    @staticmethod
    @pytest.fixture()
    def LAMMPSDATA_partial(tmpdir):
        filename = LAMMPSdata
        N_kept = 5
        u = mda.Universe(filename)
        ext = os.path.splitext(filename)[1]
        outfile = str(tmpdir.join("lammps-data-writer-test" + ext))

        with mda.Writer(outfile, n_atoms=N_kept) as w:
            w.write(u.atoms[:N_kept])

        u_new = mda.Universe(outfile)

        return u, u_new

    @pytest.mark.parametrize(
        "attr", ["masses", "charges", "velocities", "positions"]
    )
    def test_Writer_atoms(self, attr, LAMMPSDATA_partial):
        u_ref, u_new = LAMMPSDATA_partial
        if hasattr(u_ref.atoms, attr):
            assert_allclose(
                getattr(u_ref.atoms[: self.N_kept], attr),
                getattr(u_new.atoms, attr),
                err_msg="attributes different after writing",
                atol=1e-6,
            )
        else:
            with pytest.raises(AttributeError):
                getattr(u_new, attr)

    def test_n_bonds(self, LAMMPSDATA_partial):
        u_ref, u_new = LAMMPSDATA_partial
        assert len(u_new.atoms.bonds) == 4

    def test_n_angles(self, LAMMPSDATA_partial):
        u_ref, u_new = LAMMPSDATA_partial
        assert len(u_new.atoms.angles) == 4


# need more tests of the LAMMPS DCDReader


class TestLAMMPSDCDReader(RefLAMMPSDataDCD):
    flavor = "LAMMPS"

    @pytest.fixture(scope="class")
    def u(self):
        return mda.Universe(self.topology, self.trajectory, format=self.format)

    def test_Reader_is_LAMMPS(self, u):
        assert u.trajectory.flavor, self.flavor

    def get_frame_from_end(self, offset, u):
        iframe = u.trajectory.n_frames - 1 - offset
        iframe = iframe if iframe > 0 else 0
        return iframe

    def test_n_atoms(self, u):
        assert_equal(u.atoms.n_atoms, self.n_atoms)

    def test_n_frames(self, u):
        assert_equal(u.trajectory.n_frames, self.n_frames)

    def test_dimensions(self, u):
        mean_dimensions = np.mean(
            [ts.dimensions.copy() for ts in u.trajectory], axis=0
        )
        assert_allclose(mean_dimensions, self.mean_dimensions)

    def test_dt(self, u):
        assert_allclose(
            u.trajectory.dt,
            self.dt,
            err_msg="Time between frames dt is wrong.",
        )

    def test_Timestep_time(self, u):
        iframe = self.get_frame_from_end(1, u)
        assert_allclose(
            u.trajectory[iframe].time,
            iframe * self.dt,
            err_msg="Time for frame {0} (dt={1}) is wrong.".format(
                iframe, self.dt
            ),
        )

    def test_LAMMPSDCDReader_set_dt(self, u, dt=1500.0):
        u = mda.Universe(
            self.topology, self.trajectory, format=self.format, dt=dt
        )
        iframe = self.get_frame_from_end(1, u)
        assert_allclose(
            u.trajectory[iframe].time,
            iframe * dt,
            err_msg="setting time step dt={0} failed: "
            "actually used dt={1}".format(dt, u.trajectory._ts_kwargs["dt"]),
        )

    def test_wrong_time_unit(self):
        def wrong_load(unit="nm"):
            return mda.Universe(
                self.topology,
                self.trajectory,
                format=self.format,
                timeunit=unit,
            )

        with pytest.raises(TypeError):
            wrong_load()

    def test_wrong_unit(self):
        def wrong_load(unit="GARBAGE"):
            return mda.Universe(
                self.topology,
                self.trajectory,
                format=self.format,
                timeunit=unit,
            )

        with pytest.raises(ValueError):
            wrong_load()


class TestLAMMPSDCDWriter(RefLAMMPSDataDCD):
    flavor = "LAMMPS"

    @pytest.fixture(scope="class")
    def u(self):
        return mda.Universe(self.topology, self.trajectory, format=self.format)

    def test_Writer_is_LAMMPS(self, u, tmpdir):
        ext = os.path.splitext(self.trajectory)[1]
        outfile = str(tmpdir.join("lammps-writer-test" + ext))
        with mda.Writer(
            outfile, n_atoms=u.atoms.n_atoms, format=self.format
        ) as W:
            assert W.flavor, self.flavor

    def test_Writer(self, u, tmpdir, n_frames=3):
        ext = os.path.splitext(self.trajectory)[1]
        outfile = str(tmpdir.join("lammps-writer-test" + ext))

        with mda.Writer(
            outfile, n_atoms=u.atoms.n_atoms, format=self.format
        ) as w:
            for ts in u.trajectory[:n_frames]:
                w.write(u)

        short = mda.Universe(self.topology, outfile)
        assert_equal(
            short.trajectory.n_frames,
            n_frames,
            err_msg="number of frames mismatch",
        )
        assert_allclose(
            short.trajectory[n_frames - 1].positions,
            u.trajectory[n_frames - 1].positions,
            6,
            err_msg="coordinate mismatch between corresponding frames",
        )

    def test_OtherWriter_is_LAMMPS(self, u, tmpdir):
        ext = os.path.splitext(self.trajectory)[1]
        outfile = str(tmpdir.join("lammps-writer-test" + ext))
        with u.trajectory.OtherWriter(outfile) as W:
            assert W.flavor, self.flavor

    def test_OtherWriter(self, u, tmpdir):
        times = []
        ext = os.path.splitext(self.trajectory)[1]
        outfile = str(tmpdir.join("lammps-writer-test" + ext))
        with u.trajectory.OtherWriter(outfile) as w:
            for ts in u.trajectory[::-1]:
                times.append(ts.time)
                w.write(u)
        # note: the reversed trajectory records times in increasing
        #       steps, and NOT reversed, i.e. the time markers are not
        #       attached to their frames. This could be considered a bug
        #       but DCD has no way to store timestamps. Right now, we'll simply
        #       test that this is the case and pass.
        reversed = mda.Universe(self.topology, outfile)
        assert_equal(
            reversed.trajectory.n_frames,
            u.trajectory.n_frames,
            err_msg="number of frames mismatch",
        )
        rev_times = [ts.time for ts in reversed.trajectory]
        assert_allclose(
            rev_times,
            times[::-1],
            6,
            err_msg="time steps of written DCD mismatch",
        )
        assert_allclose(
            reversed.trajectory[-1].positions,
            u.trajectory[0].positions,
            6,
            err_msg="coordinate mismatch between corresponding frames",
        )


class TestLAMMPSDCDWriterClass(object):
    flavor = "LAMMPS"

    def test_Writer_is_LAMMPS(self, tmpdir):
        outfile = str(tmpdir.join("lammps-writer-test.dcd"))
        with mda.coordinates.LAMMPS.DCDWriter(outfile, n_atoms=10) as W:
            assert W.flavor, self.flavor

    def test_open(self, tmpdir):
        outfile = str(tmpdir.join("lammps-writer-test.dcd"))
        try:
            with mda.coordinates.LAMMPS.DCDWriter(outfile, n_atoms=10):
                pass
        except Exception:
            pytest.fail()

    def test_wrong_time_unit(self, tmpdir):
        outfile = str(tmpdir.join("lammps-writer-test.dcd"))
        with pytest.raises(TypeError):
            with mda.coordinates.LAMMPS.DCDWriter(
                outfile, n_atoms=10, timeunit="nm"
            ):
                pass

    def test_wrong_unit(self, tmpdir):
        outfile = str(tmpdir.join("lammps-writer-test.dcd"))
        with pytest.raises(ValueError):
            with mda.coordinates.LAMMPS.DCDWriter(
                outfile, n_atoms=10, timeunit="GARBAGE"
            ):
                pass


def test_triclinicness():
    u = mda.Universe(LAMMPScnt)

    assert u.dimensions[3] == 90.0
    assert u.dimensions[4] == 90.0
    assert u.dimensions[5] == 120.0


@pytest.fixture
def tmpout(tmpdir):
    return str(tmpdir.join("out.data"))


class TestDataWriterErrors(object):
    def test_write_no_masses(self, tmpout):
        u = make_Universe(("types",), trajectory=True)

        try:
            u.atoms.write(tmpout)
        except NoDataError as e:
            assert "masses" in e.args[0]
        else:
            pytest.fail()

    def test_write_no_types(self, tmpout):
        u = make_Universe(("masses",), trajectory=True)

        try:
            u.atoms.write(tmpout)
        except NoDataError as e:
            assert "types" in e.args[0]
        else:
            pytest.fail()

    def test_write_non_numerical_types(self, tmpout):
        u = make_Universe(("types", "masses"), trajectory=True)

        try:
            u.atoms.write(tmpout)
        except ValueError as e:
            assert "must be convertible to integers" in e.args[0]
        else:
            raise pytest.fail()


class TestLammpsDumpReader(object):
    @pytest.fixture(params=["ascii", "bz2", "gzip"])
    def u(self, tmpdir, request):
        trjtype = request.param
        if trjtype == "bz2":
            # no conversion needed
            f = LAMMPSDUMP
        else:
            f = str(tmpdir.join("lammps." + trjtype))
            with bz2.BZ2File(LAMMPSDUMP, "rb") as datain:
                data = datain.read()
            if trjtype == "ascii":
                with open(f, "wb") as fout:
                    fout.write(data)
            elif trjtype == "gzip":
                with gzip.GzipFile(f, "wb") as fout:
                    fout.write(data)

        yield mda.Universe(
            f, format="LAMMPSDUMP", lammps_coordinate_convention="auto"
        )

    @pytest.fixture()
    def u_additional_columns_true(self):
        f = LAMMPSDUMP_additional_columns
        top = LAMMPSdata_additional_columns
        return mda.Universe(
            top,
            f,
            format="LAMMPSDUMP",
            lammps_coordinate_convention="auto",
            additional_columns=True,
        )

    @pytest.fixture()
    def u_additional_columns_single(self):
        f = LAMMPSDUMP_additional_columns
        top = LAMMPSdata_additional_columns
        return mda.Universe(
            top,
            f,
            format="LAMMPSDUMP",
            lammps_coordinate_convention="auto",
            additional_columns=["q"],
        )

    @pytest.fixture()
    def u_additional_columns_multiple(self):
        f = LAMMPSDUMP_additional_columns
        top = LAMMPSdata_additional_columns
        return mda.Universe(
            top,
            f,
            format="LAMMPSDUMP",
            lammps_coordinate_convention="auto",
            additional_columns=["q", "p"],
        )

    @pytest.fixture()
    def u_additional_columns_wrong_format(self):
        f = LAMMPSDUMP_additional_columns
        top = LAMMPSdata_additional_columns
        return mda.Universe(
            top,
            f,
            format="LAMMPSDUMP",
            lammps_coordinate_convention="auto",
            additional_columns="q",
        )

    @pytest.fixture()
    def u_additional_columns_not_present(self):
        f = LAMMPSDUMP_additional_columns
        top = LAMMPSdata_additional_columns
        return mda.Universe(
            top,
            f,
            format="LAMMPSDUMP",
            lammps_coordinate_convention="auto",
            additional_columns=["q", "w"],
        )

    @pytest.fixture()
    def reference_positions(self):
        # manually copied from traj file
        data = {}

        # at timestep 500
        lo, hi = float(2.1427867124774069e-01), float(5.9857213287522608e00)
        length1 = hi - lo
        # at timestep 1000
        lo, hi = float(-5.4458069063278991e-03), float(6.2054458069063330e00)
        length2 = hi - lo
        boxes = [
            np.array([6.2, 6.2, 6.2, 90.0, 90.0, 90.0]),
            np.array([length1, length1, length1, 90.0, 90.0, 90.0]),
            np.array([length2, length2, length2, 90.0, 90.0, 90.0]),
        ]
        data["box"] = boxes
        box_mins = [
            np.array([0.0, 0.0, 0.0]),
            np.array([0.21427867, 0.21427867, 0.21427867]),
            np.array([-0.00544581, -0.00544581, -0.00544581]),
        ]
        data["mins"] = box_mins

        # data for atom id 1 in traj (ie first in frame)
        # isn't sensitive to required sorting
        atom1_pos1 = np.array([0.25, 0.25, 0.241936]) * boxes[0][:3]
        atom1_pos2 = np.array([0.278215, 0.12611, 0.322087]) * boxes[1][:3]
        atom1_pos3 = np.array([0.507123, 1.00424, 0.280972]) * boxes[2][:3]
        data["atom1_pos"] = [atom1_pos1, atom1_pos2, atom1_pos3]
        # data for atom id 13
        # *is* sensitive to reordering of positions
        # normally appears 4th in traj data
        atom13_pos1 = np.array([0.25, 0.25, 0.741936]) * boxes[0][:3]
        atom13_pos2 = np.array([0.394618, 0.263115, 0.798295]) * boxes[1][:3]
        atom13_pos3 = np.array([0.332363, 0.30544, 0.641589]) * boxes[2][:3]
        data["atom13_pos"] = [atom13_pos1, atom13_pos2, atom13_pos3]

        return data

    def test_n_atoms(self, u):
        assert len(u.atoms) == 24

    def test_length(self, u):
        assert len(u.trajectory) == 3
        for i, ts in enumerate(u.trajectory):
            assert ts.frame == i
            assert ts.data["step"] == i * 500
        for i, ts in enumerate(u.trajectory):
            assert ts.frame == i
            assert ts.data["step"] == i * 500

    def test_seeking(self, u, reference_positions):
        u.trajectory[1]

        assert_allclose(u.dimensions, reference_positions["box"][1], atol=1e-5)
        pos = (
            reference_positions["atom1_pos"][1]
            - reference_positions["mins"][1]
        )
        assert_allclose(u.atoms[0].position, pos, atol=1e-5)
        pos = (
            reference_positions["atom13_pos"][1]
            - reference_positions["mins"][1]
        )
        assert_allclose(u.atoms[12].position, pos, atol=1e-5)

    def test_boxsize(self, u, reference_positions):
        for ts, box in zip(u.trajectory, reference_positions["box"]):
            assert_allclose(ts.dimensions, box, atol=1e-5)

    def test_atom_reordering(self, u, reference_positions):
        atom1 = u.atoms[0]
        atom13 = u.atoms[12]
        for ts, atom1_pos, atom13_pos, bmin in zip(
            u.trajectory,
            reference_positions["atom1_pos"],
            reference_positions["atom13_pos"],
            reference_positions["mins"],
        ):
            assert_allclose(atom1.position, atom1_pos - bmin, atol=1e-5)
            assert_allclose(atom13.position, atom13_pos - bmin, atol=1e-5)

    @pytest.mark.parametrize(
        "system, fields",
        [
            ("u_additional_columns_true", ["q", "p"]),
            ("u_additional_columns_single", ["q"]),
            ("u_additional_columns_multiple", ["q", "p"]),
        ],
    )
    def test_additional_columns(self, system, fields, request):
        u = request.getfixturevalue(system)
        for field in fields:
            data = u.trajectory[0].data[field]
            assert_allclose(
                data, getattr(RefLAMMPSDataAdditionalColumns, field)
            )

    @pytest.mark.parametrize(
        "system",
        [
            ("u_additional_columns_wrong_format"),
        ],
    )
    def test_wrong_format_additional_colums(self, system, request):
        with pytest.raises(
            ValueError, match="Please provide an iterable containing"
        ):
            request.getfixturevalue(system)

    @pytest.mark.parametrize(
        "system",
        [
            ("u_additional_columns_not_present"),
        ],
    )
    def test_warning(self, system, request):
        with pytest.warns(match="Some of the additional"):
            request.getfixturevalue(system)


@pytest.mark.parametrize(
    "convention", ["unscaled", "unwrapped", "scaled_unwrapped"]
)
def test_open_absent_convention_fails(convention):
    with pytest.raises(ValueError, match="No coordinates following"):
        mda.Universe(
            LAMMPSDUMP,
            format="LAMMPSDUMP",
            lammps_coordinate_convention=convention,
        )


def test_open_incorrect_convention_fails():
    with pytest.raises(ValueError, match="is not a valid option"):
        mda.Universe(
            LAMMPSDUMP, format="LAMMPSDUMP", lammps_coordinate_convention="42"
        )


@pytest.mark.parametrize(
    "convention,result",
    [
        ("auto", "unscaled"),
        ("unscaled", "unscaled"),
        ("scaled", "scaled"),
        ("unwrapped", "unwrapped"),
        ("scaled_unwrapped", "scaled_unwrapped"),
    ],
)
def test_open_all_convention(convention, result):
    u = mda.Universe(
        LAMMPSDUMP_allcoords,
        format="LAMMPSDUMP",
        lammps_coordinate_convention=convention,
    )
    assert u.trajectory.lammps_coordinate_convention == result


def test_no_coordinate_info():
    with pytest.raises(ValueError, match="No coordinate information detected"):
        u = mda.Universe(
            LAMMPSDUMP_nocoords,
            format="LAMMPSDUMP",
            lammps_coordinate_convention="auto",
        )


class TestCoordinateMatches(object):
    @pytest.fixture()
    def universes(self):
        coordinate_conventions = [
            "auto",
            "unscaled",
            "scaled",
            "unwrapped",
            "scaled_unwrapped",
        ]
        universes = {
            i: mda.Universe(
                LAMMPSDUMP_allcoords,
                format="LAMMPSDUMP",
                lammps_coordinate_convention=i,
            )
            for i in coordinate_conventions
        }
        return universes

    @pytest.fixture()
    def reference_unscaled_positions(self):
        # copied from trajectory file
        # atom 340 is the first one in the trajectory so we use that
        bmin = np.array([0.02645, 0.02645, 0.02641])
        atom340_pos1_unscaled = np.array([4.48355, 0.331422, 1.59231]) - bmin
        atom340_pos2_unscaled = np.array([4.41947, 35.4403, 2.25115]) - bmin
        atom340_pos3_unscaled = np.array([4.48989, 0.360633, 2.63623]) - bmin
        return np.asarray(
            [
                atom340_pos1_unscaled,
                atom340_pos2_unscaled,
                atom340_pos3_unscaled,
            ]
        )

    def test_unscaled_reference(self, universes, reference_unscaled_positions):
        atom_340 = universes["unscaled"].atoms[339]
        for i, ts_u in enumerate(universes["unscaled"].trajectory[0:3]):
            assert_allclose(
                atom_340.position,
                reference_unscaled_positions[i, :],
                atol=1e-5,
            )

    def test_scaled_reference(self, universes, reference_unscaled_positions):
        # NOTE use of unscaled positions here due to S->R transform
        atom_340 = universes["scaled"].atoms[339]
        for i, ts_u in enumerate(universes["scaled"].trajectory[0:3]):
            assert_allclose(
                atom_340.position,
                reference_unscaled_positions[i, :],
                atol=1e-1,
            )
            # NOTE this seems a bit inaccurate?

    @pytest.fixture()
    def reference_unwrapped_positions(self):
        # copied from trajectory file
        # atom 340 is the first one in the trajectory so we use that
        atom340_pos1_unwrapped = [4.48355, 35.8378, 1.59231]
        atom340_pos2_unwrapped = [4.41947, 35.4403, 2.25115]
        atom340_pos3_unwrapped = [4.48989, 35.867, 2.63623]
        return np.asarray(
            [
                atom340_pos1_unwrapped,
                atom340_pos2_unwrapped,
                atom340_pos3_unwrapped,
            ]
        )

    def test_unwrapped_scaled_reference(
        self, universes, reference_unwrapped_positions
    ):
        atom_340 = universes["unwrapped"].atoms[339]
        for i, ts_u in enumerate(universes["unwrapped"].trajectory[0:3]):
            assert_allclose(
                atom_340.position,
                reference_unwrapped_positions[i, :],
                atol=1e-5,
            )

    def test_unwrapped_scaled_reference(
        self, universes, reference_unwrapped_positions
    ):
        # NOTE use of unscaled positions here due to S->R transform
        atom_340 = universes["scaled_unwrapped"].atoms[339]
        for i, ts_u in enumerate(
            universes["scaled_unwrapped"].trajectory[0:3]
        ):
            assert_allclose(
                atom_340.position,
                reference_unwrapped_positions[i, :],
                atol=1e-1,
            )
            # NOTE this seems a bit inaccurate?

    def test_scaled_unscaled_match(self, universes):
        assert len(universes["unscaled"].trajectory) == len(
            universes["scaled"].trajectory
        )
        for ts_u, ts_s in zip(
            universes["unscaled"].trajectory, universes["scaled"].trajectory
        ):
            assert_allclose(ts_u.positions, ts_s.positions, atol=1e-1)
            # NOTE this seems a bit inaccurate?

    def test_unwrapped_scaled_unwrapped_match(self, universes):
        assert len(universes["unwrapped"].trajectory) == len(
            universes["scaled_unwrapped"].trajectory
        )
        for ts_u, ts_s in zip(
            universes["unwrapped"].trajectory,
            universes["scaled_unwrapped"].trajectory,
        ):
            assert_allclose(ts_u.positions, ts_s.positions, atol=1e-1)
            # NOTE this seems a bit inaccurate?

    def test_auto_is_unscaled_match(self, universes):
        assert len(universes["auto"].trajectory) == len(
            universes["unscaled"].trajectory
        )
        for ts_a, ts_s in zip(
            universes["auto"].trajectory, universes["unscaled"].trajectory
        ):
            assert_allclose(ts_a.positions, ts_s.positions, atol=1e-5)


class TestLammpsTriclinic(object):
    @pytest.fixture()
    def u_dump(self):
        return mda.Universe(
            LAMMPSDUMP_triclinic,
            format="LAMMPSDUMP",
            lammps_coordinate_convention="auto",
        )

    @pytest.fixture()
    def u_data(self):
        return mda.Universe(
            LAMMPSdata_triclinic, format="data", atom_style="id type x y z"
        )

    @pytest.fixture()
    def reference_box(self):
        # manually copied from triclinic data file
        xlo = -0.32115478301032807
        xhi = 16.831069399898624
        ylo = -0.12372358703610897
        yhi = 25.95896427399614
        zlo = -0.045447071698045266
        zhi = 12.993982724334792
        xy = 1.506743915478767
        xz = -6.266414551929444
        yz = -0.42179319547892025

        box = np.zeros((3, 3), dtype=np.float64)
        box[0] = xhi - xlo, 0.0, 0.0
        box[1] = xy, yhi - ylo, 0.0
        box[2] = xz, yz, zhi - zlo

        return mda.lib.mdamath.triclinic_box(*box)

    def test_box(self, u_dump, u_data, reference_box):
        # NOTE threefold validation testing both data and dump reader.
        for ts in u_dump.trajectory:
            assert_allclose(ts.dimensions, reference_box, rtol=1e-5, atol=0)

        for ts in u_dump.trajectory:
            assert_allclose(
                ts.dimensions, u_data.dimensions, rtol=1e-5, atol=0
            )

        assert_allclose(u_data.dimensions, reference_box, rtol=1e-5, atol=0)
